from __future__ import annotations

import logging
from typing import Any, Dict, List, Optional, Tuple, Union

import networkx as nx

from woke.ir import (
    Block,
    Break,
    Continue,
    DoWhileStatement,
    ExpressionStatement,
    ForStatement,
    FunctionCall,
    FunctionDefinition,
    IfStatement,
    ModifierDefinition,
    Return,
    RevertStatement,
    StatementAbc,
    TryStatement,
    UncheckedBlock,
    WhileStatement,
)
from woke.ir.enums import GlobalSymbolsEnum
from woke.utils import StrEnum

logger = logging.getLogger(__name__)

# pyright: reportGeneralTypeIssues=false
# pyright: reportOptionalSubscript=false


class TransitionCondition(StrEnum):
    IS_TRUE = "is true"
    IS_FALSE = "is false"
    ALWAYS = "always"
    NEVER = "never"
    TRY_SUCCEEDED = "succeeded"
    TRY_REVERTED = "reverted"
    TRY_PANICKED = "panicked"
    TRY_FAILED = "failed"


class ControlFlowGraph:
    __graph: nx.DiGraph
    __declaration: Union[FunctionDefinition, ModifierDefinition]
    __statements_lookup: Dict[StatementAbc, CfgBlock]
    __start_block: CfgBlock
    __end_block: CfgBlock

    def __init__(self, declaration: Union[FunctionDefinition, ModifierDefinition]):
        if declaration.body is None:
            raise ValueError("Function body is None.")
        self.__declaration = declaration

        self.__graph = nx.DiGraph()
        self.__start_block = CfgBlock()
        self.__graph.add_node(self.__start_block)
        next_block = CfgBlock()
        self.__graph.add_node(next_block)
        self.__graph.add_edge(
            self.__start_block, next_block, condition=(TransitionCondition.ALWAYS, None)
        )
        self.__end_block = CfgBlock()
        self.__graph.add_node(self.__end_block)

        last = CfgBlock.from_statement(
            self.__graph,
            next_block,
            self.__end_block,
            None,
            None,
            declaration.body,
        )
        self.__graph.add_edge(
            last, self.__end_block, condition=(TransitionCondition.ALWAYS, None)
        )

        while _normalize(self.__graph, self.__start_block, self.__end_block):
            pass

        if len(self.__end_block.statements) != 0:
            end_block = CfgBlock()
            self.__graph.add_node(end_block)
            self.__graph.add_edge(
                self.__end_block,
                end_block,
                condition=(TransitionCondition.ALWAYS, None),
            )
            self.__end_block = end_block

        self.__statements_lookup = {
            stmt: block for block in self.__graph.nodes for stmt in block.statements
        }
        for node in self.__graph.nodes:
            for stmt in node.statements:
                self.__statements_lookup[stmt] = node
            if node.control_statement is not None:
                self.__statements_lookup[node.control_statement] = node

    @property
    def graph(self) -> nx.DiGraph:
        return self.__graph.copy(as_view=True)

    @property
    def declaration(self) -> Union[FunctionDefinition, ModifierDefinition]:
        return self.__declaration

    @property
    def start_block(self) -> CfgBlock:
        """
        Start block is guaranteed to be empty, i.e. it has no statements.
        """
        return self.__start_block

    @property
    def end_block(self) -> CfgBlock:
        """
        End block is guaranteed to be empty, i.e. it has no statements.
        """
        return self.__end_block

    def get_cfg_block(self, statement: StatementAbc) -> CfgBlock:
        return self.__statements_lookup[statement]

    def is_reachable(self, start: StatementAbc, end: StatementAbc) -> bool:
        start_block = self.__statements_lookup[start]
        end_block = self.__statements_lookup[end]
        if start_block == end_block:
            if end == end_block.control_statement:
                return True
            start_index = start_block.statements.index(start)
            end_index = end_block.statements.index(end)
            if start_index <= end_index:  # also EQUAL?
                return True
            try:
                nx.find_cycle(self.__graph, start_block)
                return True
            except nx.NetworkXNoCycle:
                return False
        else:
            return nx.has_path(self.__graph, start_block, end_block)


def _normalize(graph: nx.DiGraph, start: CfgBlock, end: CfgBlock) -> bool:
    changed = False
    to_be_removed = set()

    for node in graph.nodes:
        for out_edge in list(graph.out_edges(node, data=True)):
            if out_edge[2]["condition"][0] == TransitionCondition.NEVER:
                graph.remove_edge(out_edge[0], out_edge[1])
                changed = True

        if (
            len(node.statements) == 0
            and len(graph.out_edges(node)) == 1
            and node != start
        ):
            edge = next(iter(graph.out_edges(node, data=True)))
            if edge[2]["condition"][0] == TransitionCondition.ALWAYS:
                to = edge[1]
                in_edges = list(graph.in_edges(node, data=True))
                for from_, _, data in in_edges:
                    graph.add_edge(from_, to, condition=data["condition"])
                    graph.remove_edge(from_, node)
                to_be_removed.add(node)
                changed = True

        if (
            len(node.statements) == 0
            and len(graph.in_edges(node)) == 0
            and node not in {start, end}
        ):
            to_be_removed.add(node)
            changed = True
            for edge in list(graph.out_edges(node)):
                graph.remove_edge(edge[0], edge[1])

    for node in to_be_removed:
        graph.remove_node(node)

    return changed


class CfgBlock:
    __id_counter: int = 0
    __id: int
    __statements: List[StatementAbc]
    # control statement is always the last statement
    __control_statement: Optional[
        Union[DoWhileStatement, ForStatement, IfStatement, TryStatement, WhileStatement]
    ]

    def __init__(self):
        self.__id = self.__class__.__id_counter
        self.__class__.__id_counter += 1
        self.__statements = []
        self.__control_statement = None

    def __str__(self):
        return (
            "\n".join(statement.source for statement in self.statements)
            if len(self.statements) > 0
            else ""
        )

    @property
    def id(self) -> int:
        return self.__id

    @property
    def statements(self) -> Tuple[StatementAbc]:
        return tuple(self.__statements)

    @property
    def control_statement(
        self,
    ) -> Optional[
        Union[DoWhileStatement, ForStatement, IfStatement, TryStatement, WhileStatement]
    ]:
        return self.__control_statement

    @classmethod
    def from_statement(
        cls,
        graph: nx.DiGraph,
        prev: CfgBlock,
        function_end: CfgBlock,
        loop_body: Optional[CfgBlock],
        loop_body_next: Optional[CfgBlock],
        statement: StatementAbc,
    ) -> CfgBlock:
        if isinstance(statement, (Block, UncheckedBlock)):
            for body_statement in statement.statements:
                prev = cls.from_statement(
                    graph, prev, function_end, loop_body, loop_body_next, body_statement
                )
            return prev
        elif isinstance(statement, Break):
            prev.__statements.append(statement)
            next = CfgBlock()
            assert loop_body_next is not None
            graph.add_node(next)
            graph.add_edge(
                prev, loop_body_next, condition=(TransitionCondition.ALWAYS, None)
            )
            graph.add_edge(prev, next, condition=(TransitionCondition.NEVER, None))
            return next
        elif isinstance(statement, Continue):
            prev.__statements.append(statement)
            next = CfgBlock()
            assert loop_body is not None
            graph.add_node(next)
            graph.add_edge(
                prev, loop_body, condition=(TransitionCondition.ALWAYS, None)
            )
            graph.add_edge(prev, next, condition=(TransitionCondition.NEVER, None))
            return next
        elif isinstance(statement, DoWhileStatement):
            return cls.from_do_while_statement(graph, prev, function_end, statement)
        elif isinstance(statement, ForStatement):
            return cls.from_for_statement(graph, prev, function_end, statement)
        elif isinstance(statement, IfStatement):
            return cls.from_if_statement(
                graph, prev, function_end, loop_body, loop_body_next, statement
            )
        elif isinstance(statement, (Return, RevertStatement)):
            prev.__statements.append(statement)
            next = CfgBlock()
            graph.add_node(next)
            graph.add_edge(
                prev, function_end, condition=(TransitionCondition.ALWAYS, None)
            )
            graph.add_edge(prev, next, condition=(TransitionCondition.NEVER, None))
            return next
        elif isinstance(statement, ExpressionStatement):
            expr = statement.expression
            if (
                isinstance(expr, FunctionCall)
                and expr.function_called == GlobalSymbolsEnum.REVERT
            ):
                prev.__statements.append(statement)
                next = CfgBlock()
                graph.add_node(next)
                graph.add_edge(
                    prev, function_end, condition=(TransitionCondition.ALWAYS, None)
                )
                graph.add_edge(prev, next, condition=(TransitionCondition.NEVER, None))
                return next
            else:
                prev.__statements.append(statement)
                return prev
        elif isinstance(statement, TryStatement):
            return cls.from_try_statement(
                graph, prev, function_end, loop_body, loop_body_next, statement
            )
        elif isinstance(statement, WhileStatement):
            return cls.from_while_statement(graph, prev, function_end, statement)
        else:
            prev.__statements.append(statement)
            return prev

    @classmethod
    def from_if_statement(
        cls,
        graph: nx.DiGraph,
        prev: CfgBlock,
        function_end: CfgBlock,
        loop_body: Optional[CfgBlock],
        loop_body_next: Optional[CfgBlock],
        if_statement: IfStatement,
    ) -> CfgBlock:
        assert prev.__control_statement is None
        prev.__control_statement = if_statement
        true_block = CfgBlock()
        graph.add_node(true_block)
        true_block_end = cls.from_statement(
            graph,
            true_block,
            function_end,
            loop_body,
            loop_body_next,
            if_statement.true_body,
        )

        false_block = CfgBlock()
        graph.add_node(false_block)

        if if_statement.false_body is None:
            false_block_end = false_block
        else:
            false_block_end = cls.from_statement(
                graph,
                false_block,
                function_end,
                loop_body,
                loop_body_next,
                if_statement.false_body,
            )

        next = CfgBlock()
        graph.add_node(next)
        graph.add_edge(
            prev,
            true_block,
            condition=(TransitionCondition.IS_TRUE, if_statement.condition),
        )
        graph.add_edge(
            prev,
            false_block,
            condition=(TransitionCondition.IS_FALSE, if_statement.condition),
        )
        graph.add_edge(
            true_block_end, next, condition=(TransitionCondition.ALWAYS, None)
        )
        graph.add_edge(
            false_block_end, next, condition=(TransitionCondition.ALWAYS, None)
        )
        return next

    @classmethod
    def from_do_while_statement(
        cls,
        graph: nx.DiGraph,
        prev: CfgBlock,
        function_end: CfgBlock,
        do_while_statement: DoWhileStatement,
    ) -> CfgBlock:
        body = CfgBlock()
        graph.add_node(body)
        next = CfgBlock()
        graph.add_node(next)
        body_end = cls.from_statement(
            graph, body, function_end, body, next, do_while_statement.body
        )
        assert body_end.__control_statement is None
        body_end.__control_statement = do_while_statement

        graph.add_edge(prev, body, condition=(TransitionCondition.ALWAYS, None))
        graph.add_edge(
            body_end,
            next,
            condition=(TransitionCondition.IS_FALSE, do_while_statement.condition),
        )
        graph.add_edge(
            body_end,
            body,
            condition=(TransitionCondition.IS_TRUE, do_while_statement.condition),
        )
        return next

    @classmethod
    def from_for_statement(
        cls,
        graph: nx.DiGraph,
        prev: CfgBlock,
        function_end: CfgBlock,
        for_statement: ForStatement,
    ) -> CfgBlock:
        if for_statement.initialization_expression is not None:
            prev.__statements.append(for_statement.initialization_expression)
        assert prev.__control_statement is None
        prev.__control_statement = for_statement

        body = CfgBlock()
        graph.add_node(body)
        next = CfgBlock()
        graph.add_node(next)
        body_end = CfgBlock()
        graph.add_node(body_end)
        tmp = cls.from_statement(
            graph, body, function_end, body_end, next, for_statement.body
        )

        if tmp == body:
            body_end = tmp
            if for_statement.loop_expression is not None:
                body_end.__statements.append(for_statement.loop_expression)
        else:
            body_end.__statements = list(tmp.__statements)

            data: Any
            for start, _, data in graph.in_edges(tmp, data=True):
                graph.add_edge(start, body_end, condition=data["condition"])

            for end, _, data in graph.out_edges(tmp, data=True):
                graph.add_edge(body_end, end, condition=data["condition"])

            graph.remove_node(tmp)
            if for_statement.loop_expression is not None:
                body_end.__statements.append(for_statement.loop_expression)

        graph.add_edge(
            prev, body, condition=(TransitionCondition.IS_TRUE, for_statement.condition)
        )
        graph.add_edge(
            prev,
            next,
            condition=(TransitionCondition.IS_FALSE, for_statement.condition),
        )
        graph.add_edge(
            body_end,
            body,
            condition=(TransitionCondition.IS_TRUE, for_statement.condition),
        )
        graph.add_edge(
            body_end,
            next,
            condition=(TransitionCondition.IS_FALSE, for_statement.condition),
        )
        return next

    @classmethod
    def from_try_statement(
        cls,
        graph: nx.DiGraph,
        prev: CfgBlock,
        function_end: CfgBlock,
        loop_body: Optional[CfgBlock],
        loop_body_next: Optional[CfgBlock],
        try_statement: TryStatement,
    ) -> CfgBlock:
        assert prev.__control_statement is None
        prev.__control_statement = try_statement

        success_block = CfgBlock()
        graph.add_node(success_block)
        success_block_end = cls.from_statement(
            graph,
            success_block,
            function_end,
            loop_body,
            loop_body_next,
            try_statement.clauses[0].block,
        )

        revert_block = None
        revert_block_end = None
        panic_block = None
        panic_block_end = None
        fail_block = None
        fail_block_end = None
        for clause in try_statement.clauses[1:]:
            if clause.error_name == "Error":
                revert_block = CfgBlock()
                graph.add_node(revert_block)
                revert_block_end = cls.from_statement(
                    graph,
                    revert_block,
                    function_end,
                    loop_body,
                    loop_body_next,
                    clause.block,
                )
            elif clause.error_name == "Panic":
                panic_block = CfgBlock()
                graph.add_node(panic_block)
                panic_block_end = cls.from_statement(
                    graph,
                    panic_block,
                    function_end,
                    loop_body,
                    loop_body_next,
                    clause.block,
                )
            elif clause.error_name == "":
                fail_block = CfgBlock()
                graph.add_node(fail_block)
                fail_block_end = cls.from_statement(
                    graph,
                    fail_block,
                    function_end,
                    loop_body,
                    loop_body_next,
                    clause.block,
                )
            else:
                raise NotImplementedError(f"Unknown error name: {clause.error_name}")

        next = CfgBlock()
        graph.add_node(next)

        graph.add_edge(
            prev,
            success_block,
            condition=(TransitionCondition.TRY_SUCCEEDED, try_statement.external_call),
        )
        graph.add_edge(
            success_block_end, next, condition=(TransitionCondition.ALWAYS, None)
        )
        if revert_block is not None:
            graph.add_edge(
                prev,
                revert_block,
                condition=(
                    TransitionCondition.TRY_REVERTED,
                    try_statement.external_call,
                ),
            )
            graph.add_edge(
                revert_block_end, next, condition=(TransitionCondition.ALWAYS, None)
            )
        if panic_block is not None:
            graph.add_edge(
                prev,
                panic_block,
                condition=(
                    TransitionCondition.TRY_PANICKED,
                    try_statement.external_call,
                ),
            )
            graph.add_edge(
                panic_block_end, next, condition=(TransitionCondition.ALWAYS, None)
            )
        if fail_block is not None:
            graph.add_edge(
                prev,
                fail_block,
                condition=(TransitionCondition.TRY_FAILED, try_statement.external_call),
            )
            graph.add_edge(
                fail_block_end, next, condition=(TransitionCondition.ALWAYS, None)
            )
        return next

    @classmethod
    def from_while_statement(
        cls,
        graph: nx.DiGraph,
        prev: CfgBlock,
        function_end: CfgBlock,
        while_statement: WhileStatement,
    ) -> CfgBlock:
        assert prev.__control_statement is None
        prev.__control_statement = while_statement

        body = CfgBlock()
        graph.add_node(body)
        next = CfgBlock()
        graph.add_node(next)
        body_end = cls.from_statement(
            graph, body, function_end, body, next, while_statement.body
        )

        graph.add_edge(
            prev,
            body,
            condition=(TransitionCondition.IS_TRUE, while_statement.condition),
        )
        graph.add_edge(
            prev,
            next,
            condition=(TransitionCondition.IS_FALSE, while_statement.condition),
        )
        graph.add_edge(
            body_end,
            body,
            condition=(TransitionCondition.IS_TRUE, while_statement.condition),
        )
        graph.add_edge(
            body_end,
            next,
            condition=(TransitionCondition.IS_FALSE, while_statement.condition),
        )
        return next



import json
import platform
import subprocess
from typing import Optional

from woke.config import WokeConfig
from woke.core.solidity_version import SemanticVersion


def get_contracts_package_version(config: WokeConfig) -> Optional[SemanticVersion]:
    try:
        node_modules_path = next(
            path
            for path in config.compiler.solc.include_paths
            if "node_modules" in path.stem and path.is_dir()
        )
    except StopIteration:
        node_modules_path = config.project_root_path / "node_modules"
        if not node_modules_path.is_dir():
            return None

    try:
        out = subprocess.run(
            ["npm", "list", "@openzeppelin/contracts", "--depth=0"],
            capture_output=True,
            cwd=node_modules_path.parent,
            check=True,
            shell=(platform.system() == "Windows"),
        ).stdout.decode("utf-8")
        return SemanticVersion.fromstring(out.splitlines()[1].split("@")[-1])
    except Exception:
        return None



from __future__ import annotations

from typing import TYPE_CHECKING, Iterator, List, Tuple, Union

from woke.ir.ast import (
    SolcYulAssignment,
    SolcYulFunctionCall,
    SolcYulIdentifier,
    SolcYulLiteral,
)
from woke.ir.utils import IrInitTuple

from .abc import YulAbc, YulStatementAbc
from .function_call import YulFunctionCall
from .identifier import YulIdentifier
from .literal import YulLiteral

if TYPE_CHECKING:
    from .block import YulBlock


class YulAssignment(YulStatementAbc):
    """
    TBD
    """

    _parent: YulBlock
    _value: Union[YulFunctionCall, YulIdentifier, YulLiteral]
    _variable_names: List[YulIdentifier]

    def __init__(
        self, init: IrInitTuple, assignment: SolcYulAssignment, parent: YulAbc
    ):
        super().__init__(init, assignment, parent)
        if isinstance(assignment.value, SolcYulFunctionCall):
            self._value = YulFunctionCall(init, assignment.value, self)
        elif isinstance(assignment.value, SolcYulIdentifier):
            self._value = YulIdentifier(init, assignment.value, self)
        elif isinstance(assignment.value, SolcYulLiteral):
            self._value = YulLiteral(init, assignment.value, self)
        else:
            assert False, f"Unexpected type: {type(assignment.value)}"
        self._variable_names = [
            YulIdentifier(init, variable_name, self)
            for variable_name in assignment.variable_names
        ]

    def __iter__(self) -> Iterator[YulAbc]:
        yield self
        yield from self._value
        for variable_name in self._variable_names:
            yield from variable_name

    @property
    def parent(self) -> YulBlock:
        return self._parent

    @property
    def value(self) -> Union[YulFunctionCall, YulIdentifier, YulLiteral]:
        return self._value

    @property
    def variable_names(self) -> Tuple[YulIdentifier, ...]:
        return tuple(self._variable_names)



import logging
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional, Tuple, Union

from intervaltree import IntervalTree

from woke.ir import (
    ContractDefinition,
    DeclarationAbc,
    ErrorDefinition,
    EventDefinition,
    FunctionDefinition,
    ModifierDefinition,
    VariableDeclaration,
)
from woke.lsp.common_structures import (
    Command,
    PartialResultParams,
    Position,
    Range,
    TextDocumentIdentifier,
    TextDocumentRegistrationOptions,
    WorkDoneProgressOptions,
    WorkDoneProgressParams,
)
from woke.lsp.context import LspContext
from woke.lsp.lsp_data_model import LspModel
from woke.lsp.utils.position import changes_to_byte_offset
from woke.lsp.utils.uri import uri_to_path

logger = logging.getLogger(__name__)


class CodeLensOptions(WorkDoneProgressOptions):
    resolve_provider: Optional[bool]
    """
    Code lens has a resolve provider as well.
    """


class CodeLensRegistrationOptions(TextDocumentRegistrationOptions, CodeLensOptions):
    pass


class CodeLensParams(WorkDoneProgressParams, PartialResultParams):
    text_document: TextDocumentIdentifier
    """
    The document to request code lens for.
    """


class CodeLens(LspModel):
    """
    A code lens represents a command that should be shown along with
    source text, like the number of references, a way to run tests, etc.

    A code lens is _unresolved_ when no command is associated to it. For
    performance reasons the creation of a code lens and resolving should be done
    in two stages
    """

    range: Range
    """
    The range in which this code lens is valid. Should only span a single line.
    """
    command: Optional[Command]
    """
    The command this code lens represents.
    """
    data: Optional[Any]
    """
    A data entry field that is preserved on a code lens item between
    a code lens and a code lens resolve request.
    """


def _resolve_declaration(declaration: DeclarationAbc, context: LspContext) -> int:
    refs_count = len(declaration.references)

    if isinstance(declaration, VariableDeclaration):
        for base_function in declaration.base_functions:
            refs_count += len(base_function.references)
    elif isinstance(declaration, FunctionDefinition):
        for base_function in declaration.base_functions:
            refs_count += len(base_function.references)
        for child_function in declaration.child_functions:
            refs_count += len(child_function.references)
    elif isinstance(declaration, ModifierDefinition):
        for base_modifier in declaration.base_modifiers:
            refs_count += len(base_modifier.references)
        for child_modifier in declaration.child_modifiers:
            refs_count += len(child_modifier.references)
    return refs_count


class CodeLensCache(NamedTuple):
    original: CodeLens
    original_byte_range: Tuple[int, int]
    validity_byte_range: Tuple[int, int]


_code_lens_cache: Dict[Path, List[CodeLensCache]] = {}


def _get_code_lens_from_cache(
    context: LspContext, path: Path, forward_changes: IntervalTree
) -> Optional[List[CodeLens]]:
    if path not in _code_lens_cache:
        return None
    ret = []
    for cached_code_lens in _code_lens_cache[path]:
        changes_at_range = forward_changes[
            cached_code_lens.validity_byte_range[
                0
            ] : cached_code_lens.validity_byte_range[1]
        ]
        start, end = cached_code_lens.original_byte_range
        if len(changes_at_range) > 0:
            # changes at range, invalidate code lens
            continue
        if start == 0:
            ret.append(cached_code_lens.original)
        else:
            # recompute code lens range
            new_start = changes_to_byte_offset(forward_changes[0:start]) + start
            new_end = changes_to_byte_offset(forward_changes[0:end]) + end
            ret.append(
                CodeLens(
                    range=context.compiler.get_range_from_byte_offsets(
                        path, (new_start, new_end)
                    ),
                    command=cached_code_lens.original.command,
                    data=cached_code_lens.original.data,
                )
            )

    return ret


def _generate_code_lens(
    context: LspContext,
    path: Path,
    title: str,
    command: str,
    arguments: Optional[List],
    byte_range: Tuple[int, int],
    validity_range: Tuple[int, int],
) -> CodeLens:
    ret = CodeLens(
        range=context.compiler.get_range_from_byte_offsets(path, byte_range),
        command=Command(
            title=title,
            command=command,
            arguments=arguments,
        ),
        data=None,
    )
    _code_lens_cache[path].append(CodeLensCache(ret, byte_range, validity_range))
    return ret


async def code_lens(
    context: LspContext, params: CodeLensParams
) -> Union[None, List[CodeLens]]:
    logger.debug(f"Code lens for file {params.text_document.uri} requested")
    if not context.config.lsp.code_lens.enable:
        return None
    await context.compiler.output_ready.wait()

    path = uri_to_path(params.text_document.uri).resolve()

    if path not in context.compiler.source_units:
        forward_changes = context.compiler.get_last_compilation_forward_changes(path)
        if forward_changes is None:
            return None
        return _get_code_lens_from_cache(context, path, forward_changes)

    code_lens = []
    source_unit = context.compiler.source_units[path]

    _code_lens_cache[path] = []

    for declaration in source_unit.declarations_iter():
        refs = list(declaration.get_all_references(include_declarations=True))
        refs_count = len([ref for ref in refs if not isinstance(ref, DeclarationAbc)])
        declaration_positions = []

        for ref in refs:
            if isinstance(ref, DeclarationAbc):
                line, col = context.compiler.get_line_pos_from_byte_offset(
                    ref.file, ref.name_location[0]
                )
                declaration_positions.append(Position(line=line, character=col))

        line, col = context.compiler.get_line_pos_from_byte_offset(
            declaration.file, declaration.name_location[0]
        )

        # should include info to be able to recover command args from cache (positions could have changed)
        # but references do not work in cache mode (compilation error) anyway
        code_lens.append(
            _generate_code_lens(
                context,
                declaration.file,
                f"{refs_count} references" if refs_count != 1 else "1 reference",
                "Tools-for-Solidity.execute.references",
                [
                    params.text_document.uri,
                    Position(line=line, character=col),
                    declaration_positions,
                ],
                declaration.name_location,
                declaration.name_location,
            )
        )

        if (
            isinstance(declaration, (FunctionDefinition, ModifierDefinition))
            and declaration.implemented
        ):
            code_lens.append(
                _generate_code_lens(
                    context,
                    declaration.file,
                    "Control flow graph",
                    "Tools-for-Solidity.generate.control_flow_graph",
                    [params.text_document.uri, declaration.canonical_name],
                    declaration.name_location,
                    declaration.byte_location,
                )
            )
        elif isinstance(declaration, ContractDefinition):
            code_lens.append(
                _generate_code_lens(
                    context,
                    declaration.file,
                    "Inheritance graph",
                    "Tools-for-Solidity.generate.inheritance_graph",
                    [params.text_document.uri, declaration.canonical_name],
                    declaration.name_location,
                    declaration.name_location,
                )
            )

            code_lens.append(
                _generate_code_lens(
                    context,
                    declaration.file,
                    "Linearized inheritance graph",
                    "Tools-for-Solidity.generate.linearized_inheritance_graph",
                    [params.text_document.uri, declaration.canonical_name],
                    declaration.name_location,
                    declaration.name_location,
                )
            )

        if (
            isinstance(declaration, FunctionDefinition)
            and declaration.function_selector is not None
        ):
            code_lens.append(
                _generate_code_lens(
                    context,
                    declaration.file,
                    f"{declaration.function_selector.hex()}",
                    "Tools-for-Solidity.copy_to_clipboard",
                    [declaration.function_selector.hex()],
                    declaration.name_location,
                    declaration.byte_location,
                )
            )
        elif (
            isinstance(declaration, ErrorDefinition)
            and declaration.error_selector is not None
        ):
            code_lens.append(
                _generate_code_lens(
                    context,
                    declaration.file,
                    f"{declaration.error_selector.hex()}",
                    "Tools-for-Solidity.copy_to_clipboard",
                    [declaration.error_selector.hex()],
                    declaration.name_location,
                    declaration.byte_location,
                )
            )
        elif (
            isinstance(declaration, EventDefinition)
            and declaration.event_selector is not None
        ):
            code_lens.append(
                _generate_code_lens(
                    context,
                    declaration.file,
                    f"{declaration.event_selector.hex()}",
                    "Tools-for-Solidity.copy_to_clipboard",
                    [declaration.event_selector.hex()],
                    declaration.name_location,
                    declaration.byte_location,
                )
            )
    return code_lens



from __future__ import annotations

import asyncio
import sys
import time
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
)

import rich_click as click

if TYPE_CHECKING:
    from rich.syntax import SyntaxTheme

    from woke.config import WokeConfig
    from woke.detectors import (
        Detection,
        DetectionConfidence,
        DetectionImpact,
        Detector,
        DetectorResult,
    )


class DetectCli(click.RichGroup):  # pyright: ignore reportPrivateImportUsage
    _plugins_loaded = False
    _plugin_commands: Dict[str, click.Command] = {}

    def __init__(
        self,
        name: Optional[str] = None,
        commands: Optional[
            Union[Dict[str, click.Command], Sequence[click.Command]]
        ] = None,
        **attrs: Any,
    ):
        super().__init__(name=name, commands=commands, **attrs)

        for command in self.commands.values():
            self._inject_params(command)

    @staticmethod
    def _inject_params(command: click.Command) -> None:
        command.params.append(
            click.Argument(
                ["paths"],
                nargs=-1,  # TODO: leave this as nargs=-1? other nargs=-1 are not possible
                type=click.Path(exists=True),
            )
        )
        command.params.append(
            click.Option(
                ["--min-impact"],
                type=click.Choice(["info", "warning", "low", "medium", "high"]),
                default="info",
                help="Minimum impact level to report",
            )
        )
        command.params.append(
            click.Option(
                ["--min-confidence"],
                type=click.Choice(["low", "medium", "high"]),
                default="low",
                help="Minimum confidence level to report",
            )
        )

    def _load_plugins(self) -> None:
        # need to load the module to register detectors
        from woke.detectors import Detector

        if sys.version_info < (3, 10):
            from importlib_metadata import entry_points
        else:
            from importlib.metadata import entry_points

        detector_entry_points = entry_points().select(group="woke.plugins.detectors")
        for entry_point in detector_entry_points:
            entry_point.load()

        if (
            Path.cwd().joinpath("detectors").is_dir()
            and Path.cwd().joinpath("detectors/__init__.py").is_file()
        ):
            from importlib.util import module_from_spec, spec_from_file_location

            sys.path.insert(0, str(Path.cwd()))
            spec = spec_from_file_location("detectors", "detectors/__init__.py")
            if spec is not None and spec.loader is not None:
                module = module_from_spec(spec)
                spec.loader.exec_module(module)

        self._plugins_loaded = True

    def add_command(self, cmd: click.Command, name: Optional[str] = None) -> None:
        self._inject_params(cmd)
        super().add_command(cmd, name)

    def get_command(self, ctx: click.Context, cmd_name: str) -> Optional[click.Command]:
        if not self._plugins_loaded:
            self._load_plugins()
        return self.commands.get(cmd_name)

    def list_commands(self, ctx: click.Context) -> List[str]:
        if not self._plugins_loaded:
            self._load_plugins()
        return sorted(self.commands)

    def invoke(self, ctx: click.Context):
        ctx.obj["subcommand_args"] = ctx.args
        ctx.obj["subcommand_protected_args"] = ctx.protected_args
        super().invoke(ctx)


def _print_detection(
    detector_name: str,
    result: DetectorResult,
    config: WokeConfig,
    theme: Union[str, SyntaxTheme] = "monokai",
) -> None:
    from rich.panel import Panel
    from rich.syntax import Syntax
    from rich.tree import Tree

    from woke.detectors.api import Detection, DetectorResult

    from .console import console

    def print_result(
        info: Union[DetectorResult, Detection],
        tree: Optional[Tree],
        detector_id: Optional[str],
    ) -> Tree:
        if isinstance(info, DetectorResult):
            detection = info.detection
        else:
            detection = info

        source_unit = detection.ir_node.source_unit
        line, col = source_unit.get_line_col_from_byte_offset(
            detection.ir_node.byte_location[0]
        )
        assert source_unit._lines_index is not None
        line -= 1
        source = ""
        start_line_index = max(0, line - 3)
        end_line_index = min(len(source_unit._lines_index), line + 3)
        for i in range(start_line_index, end_line_index):
            source += source_unit._lines_index[i][0].decode("utf-8")

        link = config.general.link_format.format(
            path=source_unit.file,
            line=line + 1,
            col=col,
        )
        subtitle = f"[link={link}]{source_unit.source_unit_name}[/link]"

        title = ""
        if isinstance(info, DetectorResult):
            if info.impact == "info":
                title += "[[bold blue]INFO[/bold blue]] "
            elif info.impact == "warning":
                title += "[[bold yellow]WARNING[/bold yellow]] "
            elif info.impact == "low":
                title += "[[bold cyan]LOW[/bold cyan]] "
            elif info.impact == "medium":
                title += "[[bold magenta]MEDIUM[/bold magenta]] "
            elif info.impact == "high":
                title += "[[bold red]HIGH[/bold red]] "

        title += detection.message
        if detector_id is not None:
            title += f" \[{detector_id}]"  # pyright: ignore reportInvalidStringEscapeSequence

        panel = Panel.fit(
            Syntax(
                source,
                "solidity",
                theme=theme,
                line_numbers=True,
                start_line=start_line_index + 1,
                highlight_lines={line + 1},
            ),
            title=title,
            title_align="left",
            subtitle=subtitle,
            subtitle_align="left",
        )

        if tree is None:
            t = Tree(panel)
        else:
            t = tree.add(panel)

        for subdetection in detection.subdetections:
            print_result(subdetection, t, None)

        return t

    console.print("\n")
    tree = print_result(result, None, detector_name)
    console.print(tree)


def _filter_detections(
    detections: List[DetectorResult],
    min_confidence: DetectionConfidence,
    min_impact: DetectionImpact,
    config: WokeConfig,
) -> List[DetectorResult]:
    from woke.utils.file_utils import is_relative_to

    def _detection_ignored(detection: Detection) -> bool:
        return any(
            is_relative_to(detection.ir_node.file, p)
            for p in config.detectors.ignore_paths
        ) and all(_detection_ignored(d) for d in detection.subdetections)

    confidence_map = {
        "low": 0,
        "medium": 1,
        "high": 2,
    }
    impact_map = {
        "info": 0,
        "warning": 1,
        "low": 2,
        "medium": 3,
        "high": 4,
    }
    return [
        detection
        for detection in detections
        if confidence_map[detection.confidence] >= confidence_map[min_confidence]
        and impact_map[detection.impact] >= impact_map[min_impact]
        and not _detection_ignored(detection.detection)
    ]


@click.group(
    name="detect",
    cls=DetectCli,
    context_settings={"auto_envvar_prefix": "WOKE_DETECTOR"},
)
@click.option(
    "--no-artifacts", is_flag=True, default=False, help="Do not write build artifacts."
)
@click.pass_context
def run_detect(ctx: click.Context, no_artifacts: bool) -> None:
    """Run vulnerability detectors on the project."""

    if "--help" in ctx.obj["subcommand_args"]:
        return

    from woke.core.visitor import visit_map
    from woke.detectors.api import DetectionConfidence, DetectionImpact

    from ..compiler import SolcOutputSelectionEnum, SolidityCompiler
    from ..compiler.build_data_model import ProjectBuild
    from ..compiler.solc_frontend import SolcOutputError, SolcOutputErrorSeverityEnum
    from ..config import WokeConfig
    from ..utils import get_class_that_defined_method
    from ..utils.file_utils import is_relative_to
    from .console import console

    config = WokeConfig()
    config.load_configs()  # load ~/.woke/config.toml and ./woke.toml

    sol_files: Set[Path] = set()
    start = time.perf_counter()
    with console.status("[bold green]Searching for *.sol files...[/]"):
        for file in config.project_root_path.rglob("**/*.sol"):
            if (
                not any(
                    is_relative_to(file, p) for p in config.compiler.solc.ignore_paths
                )
                and file.is_file()
            ):
                sol_files.add(file)
    end = time.perf_counter()
    console.log(
        f"[green]Found {len(sol_files)} *.sol files in [bold green]{end - start:.2f} s[/bold green][/]"
    )

    compiler = SolidityCompiler(config)
    compiler.load(console=console)

    build: ProjectBuild
    errors: Set[SolcOutputError]
    build, errors = asyncio.run(
        compiler.compile(
            sol_files,
            [SolcOutputSelectionEnum.ALL],
            write_artifacts=not no_artifacts,
            console=console,
            no_warnings=True,
        )
    )

    errored = any(
        error.severity == SolcOutputErrorSeverityEnum.ERROR for error in errors
    )
    if errored:
        sys.exit(1)

    assert compiler.latest_build_info is not None
    assert compiler.latest_graph is not None

    detections: List[DetectorResult]
    assert isinstance(ctx.command, DetectCli)
    assert ctx.invoked_subcommand is not None
    command = ctx.command.get_command(ctx, ctx.invoked_subcommand)
    assert command is not None
    assert command.name is not None

    if hasattr(config.detectors, command.name):
        default_map = getattr(config.detectors, command.name)
    else:
        default_map = None

    cls: Type[Detector] = get_class_that_defined_method(
        command.callback
    )  # pyright: ignore reportGeneralTypeIssues
    if cls is not None:

        def _callback(*args, **kwargs):
            instance.paths = [Path(p).resolve() for p in kwargs.pop("paths", [])]

            min_confidence = DetectionConfidence(kwargs.pop("min_confidence", "low"))
            min_impact = DetectionImpact(kwargs.pop("min_impact", "low"))

            original_callback(
                instance, *args, **kwargs
            )  # pyright: ignore reportOptionalCall

            with console.status("[bold green]Running detectors...") as status:
                for path, source_unit in build.source_units.items():
                    if len(instance.paths) == 0 or any(
                        is_relative_to(path, p) for p in instance.paths
                    ):
                        status.update(
                            f"[bold green]Detecting in {source_unit.source_unit_name}..."
                        )
                        for node in source_unit:
                            visit_map[node.ast_node.node_type](instance, node)

            return _filter_detections(
                instance.detect(),
                min_confidence,
                min_impact,
                config,
            )

        instance = cls()
        instance.build = build
        instance.build_info = compiler.latest_build_info
        instance.config = config
        instance.imports_graph = (
            compiler.latest_graph.copy()
        )  # pyright: ignore reportGeneralTypeIssues

        original_callback = command.callback
        command.callback = _callback

        sub_ctx = command.make_context(
            command.name,
            [*ctx.obj["subcommand_protected_args"][1:], *ctx.obj["subcommand_args"]],
            parent=ctx,
            default_map=default_map,
        )
        with sub_ctx:
            detections = sub_ctx.command.invoke(sub_ctx)
    else:

        def _callback(*args, **kwargs):
            click.get_current_context().obj["paths"] = [
                Path(p).resolve() for p in kwargs.pop("paths", [])
            ]
            min_confidence = DetectionConfidence(kwargs.pop("min_confidence", "low"))
            min_impact = DetectionImpact(kwargs.pop("min_impact", "low"))

            return _filter_detections(
                original_callback(
                    *args, **kwargs
                ),  # pyright: ignore reportOptionalCall
                min_confidence,
                min_impact,
                config,
            )

        # this is the case without the Detector class
        args = [*ctx.obj["subcommand_protected_args"][1:], *ctx.obj["subcommand_args"]]
        ctx.obj = {
            "build": build,
            "build_info": compiler.latest_build_info,
            "config": config,
            "imports_graph": compiler.latest_graph.copy(),
        }

        if command.name != "all":
            original_callback = command.callback
            command.callback = _callback

        sub_ctx = command.make_context(
            command.name, args, parent=ctx, default_map=default_map
        )
        with sub_ctx:
            detections = sub_ctx.command.invoke(sub_ctx)

    # TODO order
    for detection in detections:
        _print_detection(ctx.invoked_subcommand, detection, config)

    if len(detections) == 0:
        console.print("No detections found")
        sys.exit(0)

    sys.exit(1)


@run_detect.command(name="all")
@click.pass_context
def run_detect_all(
    ctx: click.Context,
    paths: Tuple[str, ...],
    min_confidence: DetectionConfidence,
    min_impact: DetectionImpact,
) -> None:
    """
    Run all detectors.
    """
    from woke.core.visitor import visit_map
    from woke.utils import get_class_that_defined_method
    from woke.utils.file_utils import is_relative_to

    from .console import console

    build = ctx.obj["build"]
    config = ctx.obj["config"]
    latest_build_info = ctx.obj["build_info"]
    latest_graph = ctx.obj["imports_graph"]

    while not isinstance(ctx.command, DetectCli):
        assert ctx.parent is not None
        ctx = ctx.parent
    assert ctx.parent is not None

    all_detectors = ctx.command.list_commands(ctx)
    if config.detectors.only is None:
        only = set(all_detectors)
    else:
        only = set(config.detectors.only)

    selected_detectors = [
        d
        for d in all_detectors
        if d in only and d not in config.detectors.exclude and d != "all"
    ]
    collected_detectors: Dict[str, Detector] = {}
    detections: Dict[str, Iterable[DetectorResult]] = {}

    # TODO print enabled detectors?

    for detector_name in selected_detectors:
        assert isinstance(ctx.command, DetectCli)
        command = ctx.command.get_command(ctx, detector_name)
        assert command is not None
        assert command.name is not None

        if hasattr(config.detectors, command.name):
            default_map = getattr(config.detectors, command.name)
        else:
            default_map = None

        cls: Type[Detector] = get_class_that_defined_method(
            command.callback
        )  # pyright: ignore reportGeneralTypeIssues
        if cls is not None:

            def _callback(*args, **kwargs):  # pyright: ignore reportGeneralTypeIssues
                kwargs.pop("paths", None)
                kwargs.pop("min_confidence", None)
                kwargs.pop("min_impact", None)

                original_callback(
                    instance, *args, **kwargs
                )  # pyright: ignore reportOptionalCall

            instance = cls()
            instance.build = build
            instance.build_info = latest_build_info
            instance.config = config
            instance.imports_graph = latest_graph.copy()
            instance.paths = [Path(p).resolve() for p in paths]

            original_callback = command.callback
            command.callback = _callback

            sub_ctx = command.make_context(
                command.name,
                [
                    *ctx.parent.obj["subcommand_protected_args"][1:],
                    *ctx.parent.obj["subcommand_args"],
                ],
                parent=ctx,
                default_map=default_map,
            )
            with sub_ctx:
                sub_ctx.command.invoke(sub_ctx)

            collected_detectors[detector_name] = instance
        else:

            def _callback(*args, **kwargs):
                kwargs.pop("paths", None)
                kwargs.pop("min_confidence", None)
                kwargs.pop("min_impact", None)

                return original_callback(
                    *args, **kwargs
                )  # pyright: ignore reportOptionalCall

            # this is the case without the Detector class
            args = [
                *ctx.parent.obj["subcommand_protected_args"][1:],
                *ctx.parent.obj["subcommand_args"],
            ]
            ctx.obj = {
                "build": build,
                "build_info": latest_build_info,
                "config": config,
                "imports_graph": latest_graph.copy(),
                "paths": [Path(p).resolve() for p in paths],
            }

            original_callback = command.callback
            command.callback = _callback

            sub_ctx = command.make_context(
                command.name, args, parent=ctx, default_map=default_map
            )
            with sub_ctx:
                detections[detector_name] = _filter_detections(
                    sub_ctx.command.invoke(sub_ctx), min_confidence, min_impact, config
                )

    with console.status("[bold green]Running detectors...") as status:
        for path, source_unit in build.source_units.items():
            if len(paths) == 0 or any(is_relative_to(path, p) for p in paths):
                status.update(
                    f"[bold green]Detecting in {source_unit.source_unit_name}..."
                )
                for node in source_unit:
                    for detector in collected_detectors.values():
                        visit_map[node.ast_node.node_type](detector, node)

    for detector_name, detector in collected_detectors.items():
        detections[detector_name] = _filter_detections(
            detector.detect(), min_confidence, min_impact, config
        )

    # TODO order
    for detector_name in sorted(detections.keys()):
        for detection in detections[detector_name]:
            _print_detection(detector_name, detection, config)

    if len(detections) == 0:
        console.print("No detections found")
        sys.exit(0)

    sys.exit(1)



from __future__ import annotations

from functools import lru_cache, reduce
from operator import or_
from typing import TYPE_CHECKING, Iterator, List, Set, Tuple, Union

from woke.ir.abc import IrAbc, SolidityAbc
from woke.ir.ast import SolcUncheckedBlock
from woke.ir.enums import ModifiesStateFlag
from woke.ir.statements.abc import StatementAbc
from woke.ir.utils import IrInitTuple

if TYPE_CHECKING:
    from .block import Block
    from .do_while_statement import DoWhileStatement
    from .for_statement import ForStatement
    from .if_statement import IfStatement
    from .while_statement import WhileStatement


class UncheckedBlock(StatementAbc):
    """
    !!! example
        Lines 2-4 in the following code:
        ```solidity linenums="1"
        function inc(uint x) public pure returns(uint) {
            unchecked {
                x += 1;
            }
            return x;
        }
        ```
    """

    _ast_node: SolcUncheckedBlock
    _parent: Union[Block, DoWhileStatement, ForStatement, IfStatement, WhileStatement]

    _statements: List[StatementAbc]

    def __init__(
        self,
        init: IrInitTuple,
        unchecked_block: SolcUncheckedBlock,
        parent: SolidityAbc,
    ):
        super().__init__(init, unchecked_block, parent)
        self._statements = [
            StatementAbc.from_ast(init, statement, self)
            for statement in unchecked_block.statements
        ]

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        for statement in self._statements:
            yield from statement

    @property
    def parent(
        self,
    ) -> Union[Block, DoWhileStatement, ForStatement, IfStatement, WhileStatement]:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def statements(self) -> Tuple[StatementAbc, ...]:
        """
        Can be empty.
        Returns:
            Statements in the block.
        """
        return tuple(self._statements)

    @property
    @lru_cache(maxsize=2048)
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        return reduce(
            or_,
            (statement.modifies_state for statement in self._statements),
            set(),
        )

    def statements_iter(self) -> Iterator["StatementAbc"]:
        yield self
        for statement in self._statements:
            yield from statement.statements_iter()



from functools import lru_cache, reduce
from operator import or_
from typing import Iterator, List, Optional, Set, Tuple

from woke.ir.abc import IrAbc, SolidityAbc
from woke.ir.ast import SolcTupleExpression
from woke.ir.enums import ModifiesStateFlag
from woke.ir.expressions.abc import ExpressionAbc
from woke.ir.utils import IrInitTuple


class TupleExpression(ExpressionAbc):
    """
    TBD
    """

    _ast_node: SolcTupleExpression
    _parent: SolidityAbc  # TODO: make this more specific

    _components: List[Optional[ExpressionAbc]]
    _is_inline_array: bool

    def __init__(
        self,
        init: IrInitTuple,
        tuple_expression: SolcTupleExpression,
        parent: SolidityAbc,
    ):
        super().__init__(init, tuple_expression, parent)
        self._is_inline_array = tuple_expression.is_inline_array

        self._components = []
        for component in tuple_expression.components:
            if component is None:
                self._components.append(None)
            else:
                self._components.append(ExpressionAbc.from_ast(init, component, self))

    def __iter__(self) -> Iterator[IrAbc]:
        yield self
        for component in self._components:
            if component is not None:
                yield from component

    @property
    def parent(self) -> SolidityAbc:
        return self._parent

    @property
    def is_inline_array(self) -> bool:
        return self._is_inline_array

    @property
    def components(self) -> Tuple[Optional[ExpressionAbc], ...]:
        return tuple(self._components)

    @property
    @lru_cache(maxsize=2048)
    def is_ref_to_state_variable(self) -> bool:
        return any(
            component.is_ref_to_state_variable
            for component in self._components
            if component is not None
        )

    @property
    @lru_cache(maxsize=2048)
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        return reduce(
            or_,
            (
                component.modifies_state
                for component in self._components
                if component is not None
            ),
            set(),
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Set, Tuple, Union

from woke.ir.abc import IrAbc, SolidityAbc
from woke.ir.ast import SolcBreak
from woke.ir.enums import ModifiesStateFlag
from woke.ir.statements.abc import StatementAbc
from woke.ir.utils import IrInitTuple

if TYPE_CHECKING:
    from .block import Block
    from .do_while_statement import DoWhileStatement
    from .for_statement import ForStatement
    from .if_statement import IfStatement
    from .unchecked_block import UncheckedBlock
    from .while_statement import WhileStatement


class Break(StatementAbc):
    """
    !!! example
        `:::solidity break` in the following code:
        ```solidity
        function foo() public {
            for (uint i = 0; i < 10; i++) {
                if (i == 5)
                    break;
            }
        }
        ```
    """

    _ast_node: SolcBreak
    _parent: Union[
        Block,
        DoWhileStatement,
        ForStatement,
        IfStatement,
        UncheckedBlock,
        WhileStatement,
    ]

    def __init__(self, init: IrInitTuple, break_: SolcBreak, parent: SolidityAbc):
        super().__init__(init, break_, parent)

    @property
    def parent(
        self,
    ) -> Union[
        Block,
        DoWhileStatement,
        ForStatement,
        IfStatement,
        UncheckedBlock,
        WhileStatement,
    ]:
        """
        Returns:
            Parent IR node.
        """
        return self._parent

    @property
    def modifies_state(self) -> Set[Tuple[IrAbc, ModifiesStateFlag]]:
        return set()



from typing import Dict, List, Optional

from pydantic import BaseModel, Extra, Field, root_validator

from woke.core.enums import EvmVersionEnum
from woke.utils import StrEnum

__doc__ = """Solc standard JSON input data model as described by https://docs.soliditylang.org/en/v0.8.12/using-the-compiler.html#input-description"""


class SolcInputLanguageEnum(StrEnum):
    SOLIDITY = "Solidity"
    YUL = "Yul"


class SolcStopAfterEnum(StrEnum):
    PARSING = "parsing"


class MetadataBytecodeHashEnum(StrEnum):
    NONE = "none"
    IPFS = "ipfs"
    BZZR1 = "bzzr1"


class SolcOutputSelectionEnum(StrEnum):
    AST = "ast"
    ABI = "abi"
    DEVDOC = "devdoc"
    """Developer documentation (natspec)"""
    USERDOC = "userdoc"
    """User documentation (natspec)"""
    METADATA = "metadata"
    IR = "ir"
    """Yul intermediate representation of the code before optimization"""
    IR_OPTIMIZED = "irOptimized"
    """Intermediate representation after optimization"""
    STORAGE_LAYOUT = "storageLayout"
    """Slots, offsets and types of the contract's state variables"""
    EVM_ALL = "evm"
    """All EVM subassets"""
    EVM_ASSEMBLY = "evm.assembly"
    """New assembly format"""
    EVM_LEGACY_ASSEMBLY = "evm.legacyAssembly"
    """Old-style assembly format in JSON"""
    EVM_BYTECODE_ALL = "evm.bytecode"
    """All bytecode subassets"""
    EVM_BYTECODE_FUNCTION_DEBUG_DATA = "evm.bytecode.functionDebugData"
    """Debugging information at function level"""
    EVM_BYTECODE_OBJECT = "evm.bytecode.object"
    """Bytecode object"""
    EVM_BYTECODE_OPCODES = "evm.bytecode.opcodes"
    """Opcodes list"""
    EVM_BYTECODE_SOURCE_MAP = "evm.bytecode.sourceMap"
    """Source mapping (useful for debugging)"""
    EVM_BYTECODE_LINK_REFERENCES = "evm.bytecode.linkReferences"
    """Link references (if unlinked object)"""
    EVM_BYTECODE_GENERATED_SOURCES = "evm.bytecode.generatedSources"
    """Sources generated by the compiler"""
    EVM_DEPLOYED_BYTECODE_ALL = "evm.deployedBytecode"
    """All deployed bytecode subassets"""
    EVM_DEPLOYED_BYTECODE_FUNCTION_DEBUG_DATA = "evm.deployedBytecode.functionDebugData"
    """Debugging information at function level"""
    EVM_DEPLOYED_BYTECODE_OBJECT = "evm.deployedBytecode.object"
    """Bytecode object"""
    EVM_DEPLOYED_BYTECODE_OPCODES = "evm.deployedBytecode.opcodes"
    """Opcodes list"""
    EVM_DEPLOYED_BYTECODE_SOURCE_MAP = "evm.deployedBytecode.sourceMap"
    """Source mapping (useful for debugging)"""
    EVM_DEPLOYED_BYTECODE_LINK_REFERENCES = "evm.deployedBytecode.linkReferences"
    """Link references (if unlinked object)"""
    EVM_DEPLOYED_BYTECODE_GENERATED_SOURCES = "evm.deployedBytecode.generatedSources"
    """Sources generated by the compiler"""
    EVM_DEPLOYED_BYTECODE_IMMUTABLE_REFERENCES = (
        "evm.deployedBytecode.immutableReferences"
    )
    """Map from AST ids to bytecode ranges that reference immutables"""
    EVM_METHOD_IDENTIFIERS = "evm.methodIdentifiers"
    """The list of function hashes"""
    EVM_GAS_ESTIMATES = "evm.gasEstimates"
    """Function gas estimates"""
    EWASM_ALL = "ewasm"
    """All EWASM subassets"""
    EWASM_WAST = "ewasm.wast"
    """Ewasm in WebAssembly S-expressions format"""
    EWASM_WASM = "ewasm.wasm"
    """Ewasm in WebAssembly binary format"""
    ALL = "*"
    """All solc outputs including AST"""


class ModelCheckerEngineEnum(StrEnum):
    ALL = "all"
    BMC = "bmc"
    CHC = "chc"
    NONE = "none"


class ModelCheckerInvariantsEnum(StrEnum):
    CONTRACT = "contract"
    REENTRANCY = "reentrancy"


class ModelCheckerSolversEnum(StrEnum):
    CVC4 = "cvc4"
    SMTLIB2 = "smtlib2"
    Z3 = "z3"


class ModelCheckerTargetsEnum(StrEnum):
    CONSTANT_CONDITION = "constantCondition"
    UNDERFLOW = "underflow"
    OVERFLOW = "overflow"
    DIV_BY_ZERO = "divByZero"
    BALANCE = "balance"
    ASSERT = "assert"
    POP_EMPTY_ARRAY = "popEmptyArray"
    OUT_OF_BOUNDS = "outOfBounds"


def _to_camel(s: str) -> str:
    split = s.split("_")
    return split[0].lower() + "".join([w.capitalize() for w in split[1:]])


class SolcInputModel(BaseModel):
    class Config:
        alias_generator = _to_camel
        extra = Extra.allow


class SolcInputSource(SolcInputModel):
    keccak256: Optional[str] = None
    content: Optional[str] = None
    urls: Optional[List[str]] = None

    @root_validator
    def content_or_urls_set(cls, values):
        content, urls = values.get("content"), values.get("urls")
        assert (content is None) != (
            urls is None
        ), "SolcInputSource: exactly one of `content`, `urls` must be set"
        return values


class SolcInputOptimizerYulDetailsSettings(SolcInputModel):
    stack_allocation: Optional[bool] = None
    optimizer_steps: Optional[str] = None


class SolcInputOptimizerDetailsSettings(SolcInputModel):
    peephole: Optional[bool] = None
    inliner: Optional[bool] = None
    jumpdest_remover: Optional[bool] = None
    order_literals: Optional[bool] = None
    deduplicate: Optional[bool] = None
    cse: Optional[bool] = None
    constant_optimizer: Optional[bool] = None
    yul: Optional[bool] = None
    yul_details: Optional[SolcInputOptimizerYulDetailsSettings] = None


class SolcInputOptimizerSettings(SolcInputModel):
    enabled: Optional[bool] = None
    runs: Optional[int] = None
    details: Optional[SolcInputOptimizerDetailsSettings] = None


class SolcInputDebugRevertStringsSettingsEnum(StrEnum):
    """How to treat revert (and require) reason strings."""

    DEFAULT = "default"
    """Do not inject compiler-generated revert strings and keep user-supplied ones."""
    STRIP = "strip"
    """Remove all revert strings (if possible, i.e. if literals are used) keeping side-effects"""
    DEBUG = "debug"
    """Inject strings for compiler-generated internal reverts, implemented for ABI encoders V1 and V2 for now."""
    VERBOSE_DEBUG = "verboseDebug"
    """Even append further information to user-supplied revert strings (not yet implemented)"""


class SolcInputDebugInfoSettingsEnum(StrEnum):
    """How much extra debug information to include in comments in the produced EVM assembly and Yul code."""

    LOCATION = "location"
    """
    Annotations of the form `@src <index>:<start>:<end>` indicating the location of the corresponding element in the original Solidity file, where:
    * `<index>` is the file index matching the `@use-src` annotation,
    * `<start>` is the index of the first byte at that location,
    * `<end>` is the index of the first byte after that location.
    """
    SNIPPET = "snippet"
    """A single-line code snippet from the location indicated by `@src`. The snippet is quoted and follows the corresponding `@src` annotation."""


class SolcInputDebugSettings(SolcInputModel):
    revert_strings: Optional[SolcInputDebugRevertStringsSettingsEnum] = None
    debug_info: Optional[List[SolcInputDebugInfoSettingsEnum]] = None


class SolcInputMetadataSettings(SolcInputModel):
    use_literal_content: Optional[bool] = None
    bytecode_hash: Optional[MetadataBytecodeHashEnum] = None


class SolcInputModelCheckerSettings(SolcInputModel):
    contracts: Dict[str, List[str]] = {}  # source unit name -> list of contract names
    div_mod_with_slacks: Optional[bool] = None
    engine: Optional[ModelCheckerEngineEnum] = None
    invariants: List[ModelCheckerInvariantsEnum] = []
    show_unproved: Optional[bool] = None
    solvers: List[ModelCheckerSolversEnum] = []
    targets: List[ModelCheckerTargetsEnum] = []
    timeout: Optional[int] = None


class SolcInputSettings(SolcInputModel):
    stop_after: Optional[SolcStopAfterEnum] = None
    remappings: Optional[List[str]] = None
    optimizer: Optional[SolcInputOptimizerSettings] = None
    evm_version: Optional[EvmVersionEnum] = None
    via_IR: Optional[bool] = Field(None, alias="viaIR")
    debug: Optional[SolcInputDebugSettings] = None
    metadata: Optional[SolcInputMetadataSettings] = None
    libraries: Optional[
        Dict[str, Dict[str, str]]
    ] = None  # source unit name -> (lib name -> address)
    output_selection: Optional[
        Dict[str, Dict[str, List[SolcOutputSelectionEnum]]]
    ] = None  # source unit name -> (contract name / empty for whole file / start for all contracts ->
    model_checker: Optional[SolcInputModelCheckerSettings] = None


class SolcInput(SolcInputModel):
    language: SolcInputLanguageEnum = Field(SolcInputLanguageEnum.SOLIDITY)
    sources: Dict[str, SolcInputSource] = {}
    settings: Optional[SolcInputSettings] = None



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Callable, Dict, Union

import woke.ir as ir

if TYPE_CHECKING:
    import networkx as nx

    from woke.compiler.build_data_model import ProjectBuild, ProjectBuildInfo
    from woke.config import WokeConfig


visit_map: Dict[str, Callable] = {
    "ContractDefinition": lambda self, node: self.visit_contract_definition(node),
    "EnumDefinition": lambda self, node: self.visit_enum_definition(node),
    "EnumValue": lambda self, node: self.visit_enum_value(node),
    "ErrorDefinition": lambda self, node: self.visit_error_definition(node),
    "EventDefinition": lambda self, node: self.visit_event_definition(node),
    "FunctionDefinition": lambda self, node: self.visit_function_definition(node),
    "ModifierDefinition": lambda self, node: self.visit_modifier_definition(node),
    "StructDefinition": lambda self, node: self.visit_struct_definition(node),
    "UserDefinedValueTypeDefinition": lambda self, node: self.visit_user_defined_value_type_definition(
        node
    ),
    "VariableDeclaration": lambda self, node: self.visit_variable_declaration(node),
    "Assignment": lambda self, node: self.visit_assignment(node),
    "BinaryOperation": lambda self, node: self.visit_binary_operation(node),
    "Conditional": lambda self, node: self.visit_conditional(node),
    "ElementaryTypeNameExpression": lambda self, node: self.visit_elementary_type_name_expression(
        node
    ),
    "FunctionCall": lambda self, node: self.visit_function_call(node),
    "FunctionCallOptions": lambda self, node: self.visit_function_call_options(node),
    "Identifier": lambda self, node: self.visit_identifier(node),
    "IndexAccess": lambda self, node: self.visit_index_access(node),
    "IndexRangeAccess": lambda self, node: self.visit_index_range_access(node),
    "Literal": lambda self, node: self.visit_literal(node),
    "MemberAccess": lambda self, node: self.visit_member_access(node),
    "NewExpression": lambda self, node: self.visit_new_expression(node),
    "TupleExpression": lambda self, node: self.visit_tuple_expression(node),
    "UnaryOperation": lambda self, node: self.visit_unary_operation(node),
    "IdentifierPath": lambda self, node: self.visit_identifier_path(node),
    "ImportDirective": lambda self, node: self.visit_import_directive(node),
    "InheritanceSpecifier": lambda self, node: self.visit_inheritance_specifier(node),
    "ModifierInvocation": lambda self, node: self.visit_modifier_invocation(node),
    "OverrideSpecifier": lambda self, node: self.visit_override_specifier(node),
    "ParameterList": lambda self, node: self.visit_parameter_list(node),
    "PragmaDirective": lambda self, node: self.visit_pragma_directive(node),
    "SourceUnit": lambda self, node: self.visit_source_unit(node),
    "StructuredDocumentation": lambda self, node: self.visit_structured_documentation(
        node
    ),
    "TryCatchClause": lambda self, node: self.visit_try_catch_clause(node),
    "UsingForDirective": lambda self, node: self.visit_using_for_directive(node),
    "Block": lambda self, node: self.visit_block(node),
    "Break": lambda self, node: self.visit_break(node),
    "Continue": lambda self, node: self.visit_continue(node),
    "DoWhileStatement": lambda self, node: self.visit_do_while_statement(node),
    "EmitStatement": lambda self, node: self.visit_emit_statement(node),
    "ExpressionStatement": lambda self, node: self.visit_expression_statement(node),
    "ForStatement": lambda self, node: self.visit_for_statement(node),
    "IfStatement": lambda self, node: self.visit_if_statement(node),
    "InlineAssembly": lambda self, node: self.visit_inline_assembly(node),
    "PlaceholderStatement": lambda self, node: self.visit_placeholder_statement(node),
    "Return": lambda self, node: self.visit_return(node),
    "RevertStatement": lambda self, node: self.visit_revert_statement(node),
    "TryStatement": lambda self, node: self.visit_try_statement(node),
    "UncheckedBlock": lambda self, node: self.visit_unchecked_block(node),
    "VariableDeclarationStatement": lambda self, node: self.visit_variable_declaration_statement(
        node
    ),
    "WhileStatement": lambda self, node: self.visit_while_statement(node),
    "ArrayTypeName": lambda self, node: self.visit_array_type_name(node),
    "ElementaryTypeName": lambda self, node: self.visit_elementary_type_name(node),
    "FunctionTypeName": lambda self, node: self.visit_function_type_name(node),
    "Mapping": lambda self, node: self.visit_mapping(node),
    "UserDefinedTypeName": lambda self, node: self.visit_user_defined_type_name(node),
    "YulAssignment": lambda self, node: self.visit_yul_assignment(node),
    "YulBlock": lambda self, node: self.visit_yul_block(node),
    "YulBreak": lambda self, node: self.visit_yul_break(node),
    "YulCase": lambda self, node: self.visit_yul_case(node),
    "YulContinue": lambda self, node: self.visit_yul_continue(node),
    "YulExpressionStatement": lambda self, node: self.visit_yul_expression_statement(
        node
    ),
    "YulForLoop": lambda self, node: self.visit_yul_for_loop(node),
    "YulFunctionCall": lambda self, node: self.visit_yul_function_call(node),
    "YulFunctionDefinition": lambda self, node: self.visit_yul_function_definition(
        node
    ),
    "YulIdentifier": lambda self, node: self.visit_yul_identifier(node),
    "YulIf": lambda self, node: self.visit_yul_if(node),
    "YulLeave": lambda self, node: self.visit_yul_leave(node),
    "YulLiteral": lambda self, node: self.visit_yul_literal(node),
    "YulSwitch": lambda self, node: self.visit_yul_switch(node),
    "YulTypedName": lambda self, node: self.visit_yul_typed_name(node),
    "YulVariableDeclaration": lambda self, node: self.visit_yul_variable_declaration(
        node
    ),
}


class Visitor:
    build: ProjectBuild
    build_info: ProjectBuildInfo
    config: WokeConfig
    imports_graph: nx.DiGraph

    def visit_contract_definition(self, node: ir.ContractDefinition):
        pass

    def visit_enum_definition(self, node: ir.EnumDefinition):
        pass

    def visit_enum_value(self, node: ir.EnumValue):
        pass

    def visit_error_definition(self, node: ir.ErrorDefinition):
        pass

    def visit_event_definition(self, node: ir.EventDefinition):
        pass

    def visit_function_definition(self, node: ir.FunctionDefinition):
        pass

    def visit_modifier_definition(self, node: ir.ModifierDefinition):
        pass

    def visit_struct_definition(self, node: ir.StructDefinition):
        pass

    def visit_user_defined_value_type_definition(
        self, node: ir.UserDefinedValueTypeDefinition
    ):
        pass

    def visit_variable_declaration(self, node: ir.VariableDeclaration):
        pass

    # expressions
    def visit_assignment(self, node: ir.Assignment):
        pass

    def visit_binary_operation(self, node: ir.BinaryOperation):
        pass

    def visit_conditional(self, node: ir.Conditional):
        pass

    def visit_elementary_type_name_expression(
        self, node: ir.ElementaryTypeNameExpression
    ):
        pass

    def visit_function_call(self, node: ir.FunctionCall):
        pass

    def visit_function_call_options(self, node: ir.FunctionCallOptions):
        pass

    def visit_identifier(self, node: ir.Identifier):
        pass

    def visit_index_access(self, node: ir.IndexAccess):
        pass

    def visit_index_range_access(self, node: ir.IndexRangeAccess):
        pass

    def visit_literal(self, node: ir.Literal):
        pass

    def visit_member_access(self, node: ir.MemberAccess):
        pass

    def visit_new_expression(self, node: ir.NewExpression):
        pass

    def visit_tuple_expression(self, node: ir.TupleExpression):
        pass

    def visit_unary_operation(self, node: ir.UnaryOperation):
        pass

    # meta
    def visit_identifier_path(self, node: ir.IdentifierPath):
        pass

    def visit_import_directive(self, node: ir.ImportDirective):
        pass

    def visit_inheritance_specifier(self, node: ir.InheritanceSpecifier):
        pass

    def visit_modifier_invocation(self, node: ir.ModifierInvocation):
        pass

    def visit_override_specifier(self, node: ir.OverrideSpecifier):
        pass

    def visit_parameter_list(self, node: ir.ParameterList):
        pass

    def visit_pragma_directive(self, node: ir.PragmaDirective):
        pass

    def visit_source_unit(self, node: ir.SourceUnit):
        pass

    def visit_structured_documentation(self, node: ir.StructuredDocumentation):
        pass

    def visit_try_catch_clause(self, node: ir.TryCatchClause):
        pass

    def visit_using_for_directive(self, node: ir.UsingForDirective):
        pass

    # statements
    def visit_block(self, node: ir.Block):
        pass

    def visit_break(self, node: ir.Break):
        pass

    def visit_continue(self, node: ir.Continue):
        pass

    def visit_do_while_statement(self, node: ir.DoWhileStatement):
        pass

    def visit_emit_statement(self, node: ir.EmitStatement):
        pass

    def visit_expression_statement(self, node: ir.ExpressionStatement):
        pass

    def visit_for_statement(self, node: ir.ForStatement):
        pass

    def visit_if_statement(self, node: ir.IfStatement):
        pass

    def visit_inline_assembly(self, node: ir.InlineAssembly):
        pass

    def visit_placeholder_statement(self, node: ir.PlaceholderStatement):
        pass

    def visit_return(self, node: ir.Return):
        pass

    def visit_revert_statement(self, node: ir.RevertStatement):
        pass

    def visit_try_statement(self, node: ir.TryStatement):
        pass

    def visit_unchecked_block(self, node: ir.UncheckedBlock):
        pass

    def visit_variable_declaration_statement(
        self, node: ir.VariableDeclarationStatement
    ):
        pass

    def visit_while_statement(self, node: ir.WhileStatement):
        pass

    # type names
    def visit_array_type_name(self, node: ir.ArrayTypeName):
        pass

    def visit_elementary_type_name(self, node: ir.ElementaryTypeName):
        pass

    def visit_function_type_name(self, node: ir.FunctionTypeName):
        pass

    def visit_mapping(self, node: ir.Mapping):
        pass

    def visit_user_defined_type_name(self, node: ir.UserDefinedTypeName):
        pass

    # yul
    def visit_yul_assignment(self, node: ir.YulAssignment):
        pass

    def visit_yul_block(self, node: ir.YulBlock):
        pass

    def visit_yul_break(self, node: ir.YulBreak):
        pass

    def visit_yul_case(self, node: ir.YulCase):
        pass

    def visit_yul_continue(self, node: ir.YulContinue):
        pass

    def visit_yul_expression_statement(self, node: ir.YulExpressionStatement):
        pass

    def visit_yul_for_loop(self, node: ir.YulForLoop):
        pass

    def visit_yul_function_call(self, node: ir.YulFunctionCall):
        pass

    def visit_yul_function_definition(self, node: ir.YulFunctionDefinition):
        pass

    def visit_yul_identifier(self, node: ir.YulIdentifier):
        pass

    def visit_yul_if(self, node: ir.YulIf):
        pass

    def visit_yul_leave(self, node: ir.YulLeave):
        pass

    def visit_yul_literal(self, node: ir.YulLiteral):
        pass

    def visit_yul_switch(self, node: ir.YulSwitch):
        pass

    def visit_yul_typed_name(self, node: ir.YulTypedName):
        pass

    def visit_yul_variable_declaration(self, node: ir.YulVariableDeclaration):
        pass

    def generate_link_from_line_col(
        self, path: Union[str, Path], line: int, col: int
    ) -> str:
        if isinstance(path, Path):
            path = str(path.resolve())
        return self.config.general.link_format.format(path=path, line=line, col=col)

    def generate_link(self, node: ir.IrAbc) -> str:
        if isinstance(node, ir.DeclarationAbc):
            line, col = node.source_unit.get_line_col_from_byte_offset(
                node.name_location[0]
            )
        else:
            line, col = node.source_unit.get_line_col_from_byte_offset(
                node.byte_location[0]
            )
        return self.generate_link_from_line_col(node.file, line, col)



import re
from dataclasses import dataclass
from typing import Dict, List, Optional, Union

from pydantic import BaseModel, Field, StrictBool, StrictInt, StrictStr
from pydantic.class_validators import root_validator
from typing_extensions import Annotated, Literal

from .enums import *

REGEX_SRC = re.compile(r"(-?\d+):(-?\d+):(-?\d+)")
PYDANTIC_CONFIG_EXTRA = "forbid"
PYDANTIC_CONFIG_ALLOW_MUTATION = False


def to_camel(s: str) -> str:
    split = s.split("_")
    return split[0].lower() + "".join([w.capitalize() for w in split[1:]])


class AstModel(BaseModel):
    class Config:
        alias_generator = to_camel
        extra = PYDANTIC_CONFIG_EXTRA
        allow_mutation = PYDANTIC_CONFIG_ALLOW_MUTATION


SolcTopLevelMemberUnion = Annotated[
    Union[
        "SolcPragmaDirective",
        "SolcImportDirective",
        # new in solc X
        "SolcVariableDeclaration",
        "SolcEnumDefinition",
        "SolcFunctionDefinition",
        "SolcStructDefinition",
        "SolcErrorDefinition",
        # new in solc 0.8.8
        "SolcUserDefinedValueTypeDefinition",
        # new in solc 0.8.13
        "SolcUsingForDirective",
        # everywhere
        "SolcContractDefinition",
    ],
    Field(discriminator="node_type"),
]

SolcContractMemberUnion = Annotated[
    Union[
        "SolcEnumDefinition",
        "SolcErrorDefinition",
        "SolcEventDefinition",
        "SolcFunctionDefinition",
        "SolcModifierDefinition",
        "SolcStructDefinition",
        "SolcUserDefinedValueTypeDefinition",
        "SolcUsingForDirective",
        "SolcVariableDeclaration",
    ],
    Field(discriminator="node_type"),
]

SolcLibraryNameUnion = Annotated[
    Union[
        "SolcUserDefinedTypeName",
        "SolcIdentifierPath",
    ],
    Field(discriminator="node_type"),
]

OptionalSolcLibraryNameUnion = Annotated[
    Union["SolcUserDefinedTypeName", "SolcIdentifierPath", None],
    Field(discriminator="node_type"),
]

SolcTypeNameUnion = Annotated[
    Union[
        "SolcArrayTypeName",
        "SolcElementaryTypeName",
        "SolcFunctionTypeName",
        "SolcMapping",
        "SolcUserDefinedTypeName",
    ],
    Field(discriminator="node_type"),
]

OptionalSolcTypeNameUnion = Annotated[
    Union[
        "SolcArrayTypeName",
        "SolcElementaryTypeName",
        "SolcFunctionTypeName",
        "SolcMapping",
        "SolcUserDefinedTypeName",
        None,
    ],
    Field(discriminator="node_type"),
]

SolcExpressionUnion = Annotated[
    Union[
        "SolcAssignment",
        "SolcBinaryOperation",
        "SolcConditional",
        "SolcElementaryTypeNameExpression",
        "SolcFunctionCall",
        "SolcFunctionCallOptions",
        "SolcIdentifier",
        "SolcIndexAccess",
        "SolcIndexRangeAccess",
        "SolcLiteral",
        "SolcMemberAccess",
        "SolcNewExpression",
        "SolcTupleExpression",
        "SolcUnaryOperation",
    ],
    Field(discriminator="node_type"),
]

OptionalSolcExpressionUnion = Annotated[
    Union[
        "SolcAssignment",
        "SolcBinaryOperation",
        "SolcConditional",
        "SolcElementaryTypeNameExpression",
        "SolcFunctionCall",
        "SolcFunctionCallOptions",
        "SolcIdentifier",
        "SolcIndexAccess",
        "SolcIndexRangeAccess",
        "SolcLiteral",
        "SolcMemberAccess",
        "SolcNewExpression",
        "SolcTupleExpression",
        "SolcUnaryOperation",
        None,
    ],
    Field(discriminator="node_type"),
]

SolcInitExprUnion = Annotated[
    Union[
        "SolcExpressionStatement",
        "SolcVariableDeclarationStatement",
    ],
    Field(discriminator="node_type"),
]

OptionalSolcInitExprUnion = Annotated[
    Union[
        "SolcExpressionStatement",
        "SolcVariableDeclarationStatement",
        None,
    ],
    Field(discriminator="node_type"),
]

SolcStatementUnion = Annotated[
    Union[
        "SolcBlock",
        "SolcBreak",
        "SolcContinue",
        "SolcDoWhileStatement",
        "SolcEmitStatement",
        "SolcExpressionStatement",
        "SolcForStatement",
        "SolcIfStatement",
        "SolcInlineAssembly",
        "SolcPlaceholderStatement",
        "SolcReturn",
        "SolcRevertStatement",
        "SolcTryStatement",
        "SolcUncheckedBlock",
        "SolcVariableDeclarationStatement",
        "SolcWhileStatement",
    ],
    Field(discriminator="node_type"),
]

OptionalSolcStatementUnion = Annotated[
    Union[
        "SolcBlock",
        "SolcBreak",
        "SolcContinue",
        "SolcDoWhileStatement",
        "SolcEmitStatement",
        "SolcExpressionStatement",
        "SolcForStatement",
        "SolcIfStatement",
        "SolcInlineAssembly",
        "SolcPlaceholderStatement",
        "SolcReturn",
        "SolcRevertStatement",
        "SolcTryStatement",
        "SolcUncheckedBlock",
        "SolcVariableDeclarationStatement",
        "SolcWhileStatement",
        None,
    ],
    Field(discriminator="node_type"),
]

SolcYulStatementUnion = Annotated[
    Union[
        "SolcYulAssignment",
        "SolcYulBlock",
        "SolcYulBreak",
        "SolcYulContinue",
        "SolcYulExpressionStatement",
        "SolcYulLeave",
        "SolcYulForLoop",
        "SolcYulFunctionDefinition",
        "SolcYulIf",
        "SolcYulSwitch",
        "SolcYulVariableDeclaration",
    ],
    Field(discriminator="node_type"),
]

SolcYulExpressionUnion = Annotated[
    Union[
        "SolcYulFunctionCall",
        "SolcYulIdentifier",
        "SolcYulLiteral",
    ],
    Field(discriminator="node_type"),
]

OptionalSolcYulExpressionUnion = Annotated[
    Union[
        "SolcYulFunctionCall",
        "SolcYulIdentifier",
        "SolcYulLiteral",
        None,
    ],
    Field(discriminator="node_type"),
]

# ModifierInvocation
ModifierName = Annotated[
    Union["SolcIdentifier", "SolcIdentifierPath"],
    Field(discriminator="node_type"),
]

# InheritanceSpecifier
BaseName = Annotated[
    Union["SolcUserDefinedTypeName", "SolcIdentifierPath"],
    Field(discriminator="node_type"),
]

Override = Annotated[
    Union[
        "SolcUserDefinedTypeName",
        "SolcIdentifierPath",
    ],
    Field(discriminator="node_type"),
]


class AstNodeId(int):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if not isinstance(v, int):
            raise TypeError(f"{cls.__name__} must be an int")
        return v

    def __repr__(self):
        return f"AstNodeId({int(self)})"


@dataclass
class Src:
    byte_offset: int
    byte_length: int
    file_id: int

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if not isinstance(v, str):
            raise TypeError(f"{cls.__name__} must be a str")
        match = re.search(REGEX_SRC, v)
        assert (
            match
        ), f"Src must be in the format '<byte_offset>:<byte_length>:<file_id>': {v}"
        [m1, m2, m3] = [int(match.group(i)) for i in range(1, 4)]
        return Src(byte_offset=m1, byte_length=m2, file_id=m3)


class TypeDescriptionsModel(AstModel):
    type_identifier: Optional[StrictStr]
    type_string: Optional[StrictStr]


class SymbolAliasModel(AstModel):  # helper class
    foreign: "SolcIdentifier"
    local: Optional[StrictStr]
    name_location: Optional[Src]  # new in 0.8.2


# InlineAssembly
class ExternalReferenceModel(AstModel):  # helper class
    declaration: AstNodeId
    is_offset: StrictBool
    is_slot: StrictBool
    src: Src
    value_size: StrictInt
    suffix: Optional[InlineAssemblySuffix]


class SolcNode(AstModel):
    node_type: str
    src: Src


class SolidityNode(SolcNode):
    id: AstNodeId

    def __iter__(self):
        def iter_list(l: List):
            for item in l:
                if isinstance(item, SolidityNode):
                    yield item
                    yield from item
                elif isinstance(item, List):
                    yield from iter_list(item)

        for item in self.__dict__.values():
            if isinstance(item, SolidityNode):
                yield item
                yield from item
            elif isinstance(item, List):
                yield from iter_list(item)


class YulNode(SolcNode):
    pass


class SolcSourceUnit(SolidityNode):
    # override alias
    node_type: Literal["SourceUnit"] = Field(alias="nodeType")
    # required
    absolute_path: StrictStr
    exported_symbols: Dict[StrictStr, List[AstNodeId]]
    nodes: List[SolcTopLevelMemberUnion]
    # optional
    license: Optional[StrictStr]


# todo: replace by __root__
AstSolc = SolcSourceUnit


class SolcPragmaDirective(SolidityNode):
    # override alias
    node_type: Literal["PragmaDirective"] = Field(alias="nodeType")
    literals: List[StrictStr]


class SolcImportDirective(SolidityNode):
    # override alias
    node_type: Literal["ImportDirective"] = Field(alias="nodeType")
    # required
    absolute_path: StrictStr
    file: StrictStr
    scope: AstNodeId
    source_unit: AstNodeId
    symbol_aliases: List[SymbolAliasModel]
    unit_alias: StrictStr
    # optional
    name_location: Optional[Src]  # new in 0.8.2

    def __iter__(self):
        for symbol_alias in self.symbol_aliases:
            yield symbol_alias.foreign
            yield from symbol_alias.foreign


class SolcVariableDeclaration(SolidityNode):
    # required
    # override alias
    node_type: Literal["VariableDeclaration"] = Field(alias="nodeType")
    name: StrictStr
    constant: StrictBool
    scope: AstNodeId
    state_variable: StrictBool
    storage_location: DataLocation
    type_descriptions: "TypeDescriptionsModel"
    visibility: Visibility
    # optional
    name_location: Optional[Src]  # new in 0.8.2
    # immutable is new in 0.6.5 but `mutability` field is set in >=0.6.6
    # in 0.6.5 `mutability` field is not exported for immutable variables because of a bug
    # constant variables were distinguished by `constant` field in versions <= 0.6.5 (this field is still present in newer versions)
    mutability: Optional[Mutability]
    base_functions: Optional[List[AstNodeId]]
    documentation: Optional[
        "SolcStructuredDocumentation"
    ]  # added in 0.6.9 for state variables
    function_selector: Optional[StrictStr]
    indexed: Optional[StrictBool]
    overrides: Optional["SolcOverrideSpecifier"]
    type_name: OptionalSolcTypeNameUnion  # is None only for <0.5.0 where `var` keyword was supported
    value: OptionalSolcExpressionUnion


class SolcEnumDefinition(SolidityNode):
    # override alias
    node_type: Literal["EnumDefinition"] = Field(alias="nodeType")
    # required
    name: StrictStr
    canonical_name: StrictStr
    members: List["SolcEnumValue"]
    # optional
    name_location: Optional[Src]  # new in 0.8.2
    documentation: Optional["SolcStructuredDocumentation"]  # new in 0.8.20


class SolcFunctionDefinition(SolidityNode):
    # override alias
    node_type: Literal["FunctionDefinition"] = Field(alias="nodeType")
    # required
    name: StrictStr
    implemented: StrictBool
    kind: FunctionKind
    modifiers: List["SolcModifierInvocation"]
    parameters: "SolcParameterList"
    return_parameters: "SolcParameterList"
    scope: AstNodeId
    state_mutability: StateMutability
    virtual: StrictBool
    visibility: Visibility
    # optional
    name_location: Optional[Src]  # new in 0.8.2
    base_functions: Optional[List[AstNodeId]]
    documentation: Optional[Union["SolcStructuredDocumentation", str]]
    function_selector: Optional[StrictStr]
    body: Optional["SolcBlock"]
    overrides: Optional["SolcOverrideSpecifier"]


class SolcStructDefinition(SolidityNode):
    # override alias
    node_type: Literal["StructDefinition"] = Field(alias="nodeType")
    # required
    name: StrictStr
    canonical_name: StrictStr
    members: List["SolcVariableDeclaration"]
    scope: AstNodeId
    visibility: Visibility
    # optional
    name_location: Optional[Src]  # new in 0.8.2
    documentation: Optional["SolcStructuredDocumentation"]  # new in 0.8.20


# new in 0.8.4
class SolcErrorDefinition(SolidityNode):
    # override alias
    node_type: Literal["ErrorDefinition"] = Field(alias="nodeType")
    # required
    name: StrictStr
    name_location: Src
    parameters: "SolcParameterList"
    # optional
    documentation: Optional["SolcStructuredDocumentation"]
    error_selector: Optional[StrictStr]


# new in 0.8.0
class SolcUserDefinedValueTypeDefinition(SolidityNode):
    # override alias
    node_type: Literal["UserDefinedValueTypeDefinition"] = Field(alias="nodeType")
    # required
    name: StrictStr
    underlying_type: "SolcElementaryTypeName"
    # optional
    name_location: Optional[Src]  # new in 0.8.2
    canonical_name: Optional[
        StrictStr
    ]  # should be present but because of a bug it is exported in >=0.8.9


class SolcContractDefinition(SolidityNode):
    # override alias
    node_type: Literal["ContractDefinition"] = Field(alias="nodeType")
    # required
    name: StrictStr
    abstract: StrictBool
    base_contracts: List["SolcInheritanceSpecifier"]
    contract_dependencies: List[AstNodeId]
    contract_kind: ContractKind
    linearized_base_contracts: List[AstNodeId]
    nodes: List[SolcContractMemberUnion]
    scope: AstNodeId
    # optional
    name_location: Optional[Src]  # new in 0.8.2
    canonical_name: Optional[
        StrictStr
    ]  # should be present but because of a bug it is exported in >=0.8.9
    fully_implemented: Optional[
        StrictBool
    ]  # missing when a file that imports the contract cannot be compiled
    documentation: Optional[Union["SolcStructuredDocumentation", str]]
    used_errors: Optional[List[AstNodeId]]  # new in 0.8.4
    used_events: Optional[List[AstNodeId]]  # new in 0.8.20
    internal_function_ids: Optional[Dict[int, StrictInt]] = Field(
        alias="internalFunctionIDs"
    )


class SolcEventDefinition(SolidityNode):
    # override alias
    node_type: Literal["EventDefinition"] = Field(alias="nodeType")
    # required
    name: StrictStr
    anonymous: StrictBool
    parameters: "SolcParameterList"
    # optional
    name_location: Optional[Src]  # new in 0.8.2
    documentation: Optional[Union["SolcStructuredDocumentation", str]]
    event_selector: Optional[
        StrictStr
    ]  # an example: "0d3648bd0f6ba80134a33ba9275ac585d9d315f0ad8355cddefde31afa28d0e9"


class SolcModifierDefinition(SolidityNode):
    # override alias
    node_type: Literal["ModifierDefinition"] = Field(alias="nodeType")
    # required
    name: StrictStr
    parameters: "SolcParameterList"
    virtual: StrictBool
    visibility: Visibility
    # optional
    body: Optional["SolcBlock"]
    name_location: Optional[Src]  # new in 0.8.2
    base_modifiers: Optional[List[AstNodeId]]
    documentation: Optional[Union["SolcStructuredDocumentation", str]]
    overrides: Optional["SolcOverrideSpecifier"]


class UsingForDirectiveFunction(AstModel):
    function: Optional["SolcIdentifierPath"]
    definition: Optional["SolcIdentifierPath"]  # added in 0.8.19
    operator: Optional[Union[UnaryOpOperator, BinaryOpOperator]]  # added in 0.8.19


class SolcUsingForDirective(SolidityNode):
    # override alias
    node_type: Literal["UsingForDirective"] = Field(alias="nodeType")
    # required
    # optional
    function_list: Optional[List[UsingForDirectiveFunction]]
    library_name: OptionalSolcLibraryNameUnion
    type_name: OptionalSolcTypeNameUnion
    global_: Optional[StrictBool] = Field(alias="global")

    def __iter__(self):
        if self.function_list is not None:
            for function in self.function_list:
                if function.function is not None:
                    yield function.function
                    yield from function.function
                if function.definition is not None:
                    yield function.definition
                    yield from function.definition
        if self.library_name is not None:
            yield self.library_name
            yield from self.library_name
        if self.type_name is not None:
            yield self.type_name
            yield from self.type_name


class SolcArrayTypeName(SolidityNode):
    # override alias
    node_type: Literal["ArrayTypeName"] = Field(alias="nodeType")
    # required
    type_descriptions: TypeDescriptionsModel
    base_type: SolcTypeNameUnion
    # optional
    length: OptionalSolcExpressionUnion


class SolcElementaryTypeName(SolidityNode):
    # override alias
    node_type: Literal["ElementaryTypeName"] = Field(alias="nodeType")
    # required
    type_descriptions: TypeDescriptionsModel
    name: StrictStr
    # optional
    state_mutability: Optional[StateMutability]


class SolcFunctionTypeName(SolidityNode):
    # override alias
    node_type: Literal["FunctionTypeName"] = Field(alias="nodeType")
    # required
    type_descriptions: TypeDescriptionsModel
    parameter_types: "SolcParameterList"
    return_parameter_types: "SolcParameterList"
    state_mutability: StateMutability
    visibility: Visibility
    # optional


class SolcMapping(SolidityNode):
    # override alias
    node_type: Literal["Mapping"] = Field(alias="nodeType")
    # required
    type_descriptions: TypeDescriptionsModel
    key_type: SolcTypeNameUnion
    value_type: SolcTypeNameUnion
    # optional
    key_name: Optional[StrictStr]  # new in 0.8.18
    key_name_location: Optional[Src]  # new in 0.8.18
    value_name: Optional[StrictStr]  # new in 0.8.18
    value_name_location: Optional[Src]  # new in 0.8.18


class SolcUserDefinedTypeName(SolidityNode):
    node_type: Literal["UserDefinedTypeName"] = Field(alias="nodeType")
    # required
    type_descriptions: TypeDescriptionsModel
    referenced_declaration: AstNodeId
    # optional
    contract_scope: Optional[AstNodeId]  # removed in 0.8.0
    name: Optional[StrictStr]  # removed in 0.8.0 in favor of path_node
    path_node: Optional["SolcIdentifierPath"]  # added in 0.8.0


class SolcBlock(SolidityNode):
    # override alias
    node_type: Literal["Block"] = Field(alias="nodeType")
    statements: List[SolcStatementUnion]
    # required
    # optional
    documentation: Optional[StrictStr]


class SolcBreak(SolidityNode):
    # override alias
    node_type: Literal["Break"] = Field(alias="nodeType")
    # required
    # optional
    documentation: Optional[StrictStr]


class SolcContinue(SolidityNode):
    # override alias
    node_type: Literal["Continue"] = Field(alias="nodeType")
    # required
    # optional
    documentation: Optional[StrictStr]


class SolcDoWhileStatement(SolidityNode):
    # override alias
    node_type: Literal["DoWhileStatement"] = Field(alias="nodeType")
    # required
    body: SolcStatementUnion
    condition: SolcExpressionUnion
    # optional
    documentation: Optional[StrictStr]


class SolcEmitStatement(SolidityNode):
    # override alias
    node_type: Literal["EmitStatement"] = Field(alias="nodeType")
    # required
    event_call: "SolcFunctionCall"
    # optional
    documentation: Optional[StrictStr]


class SolcExpressionStatement(SolidityNode):
    # override alias
    node_type: Literal["ExpressionStatement"] = Field(alias="nodeType")
    # required
    expression: SolcExpressionUnion
    # optional
    documentation: Optional[StrictStr]


class SolcForStatement(SolidityNode):
    # override alias
    node_type: Literal["ForStatement"] = Field(alias="nodeType")
    # required
    body: SolcStatementUnion
    # optional
    documentation: Optional[StrictStr]
    condition: OptionalSolcExpressionUnion
    initialization_expression: OptionalSolcInitExprUnion
    loop_expression: Optional[SolcExpressionStatement]


class SolcIfStatement(SolidityNode):
    # override alias
    node_type: Literal["IfStatement"] = Field(alias="nodeType")
    # required
    condition: SolcExpressionUnion
    true_body: SolcStatementUnion
    # optional
    documentation: Optional[StrictStr]
    false_body: OptionalSolcStatementUnion


class SolcInlineAssembly(SolidityNode):
    # override alias
    node_type: Literal["InlineAssembly"] = Field(alias="nodeType")
    # required
    # this one requires special care...
    ast: "SolcYulBlock" = Field(..., alias="AST")
    evm_version: InlineAssemblyEvmVersion
    external_references: List[ExternalReferenceModel]
    # optional
    documentation: Optional[StrictStr]
    flags: Optional[List[InlineAssemblyFlag]]


class SolcPlaceholderStatement(SolidityNode):
    # override alias
    node_type: Literal["PlaceholderStatement"] = Field(alias="nodeType")
    # required
    # optional
    documentation: Optional[StrictStr]


class SolcReturn(SolidityNode):
    # override alias
    node_type: Literal["Return"] = Field(alias="nodeType")
    # required
    # optional
    function_return_parameters: Optional[AstNodeId]
    documentation: Optional[StrictStr]
    expression: OptionalSolcExpressionUnion


class SolcRevertStatement(SolidityNode):
    # override alias
    node_type: Literal["RevertStatement"] = Field(alias="nodeType")
    # required
    error_call: "SolcFunctionCall"
    # optional
    documentation: Optional[StrictStr]


class SolcTryStatement(SolidityNode):
    # override alias
    node_type: Literal["TryStatement"] = Field(alias="nodeType")
    # required
    clauses: List["SolcTryCatchClause"]
    external_call: "SolcFunctionCall"
    # optional
    documentation: Optional[StrictStr]


# new in 0.8.0
class SolcUncheckedBlock(SolidityNode):
    # override alias
    node_type: Literal["UncheckedBlock"] = Field(alias="nodeType")
    # required
    statements: List[SolcStatementUnion]
    # optional
    documentation: Optional[StrictStr]


class SolcVariableDeclarationStatement(SolidityNode):
    # override alias
    node_type: Literal["VariableDeclarationStatement"] = Field(alias="nodeType")
    # required
    assignments: List[Optional[AstNodeId]]
    declarations: List[Optional["SolcVariableDeclaration"]]
    # optional
    documentation: Optional[StrictStr]
    initial_value: OptionalSolcExpressionUnion


class SolcWhileStatement(SolidityNode):
    # override alias
    node_type: Literal["WhileStatement"] = Field(alias="nodeType")
    # required
    body: SolcStatementUnion
    condition: SolcExpressionUnion
    # optional
    documentation: Optional[StrictStr]


class SolcAssignment(SolidityNode):
    # override alias
    node_type: Literal["Assignment"] = Field(alias="nodeType")
    # required
    is_constant: StrictBool
    is_l_value: StrictBool
    is_pure: StrictBool
    l_value_requested: StrictBool
    type_descriptions: TypeDescriptionsModel
    left_hand_side: SolcExpressionUnion
    operator: AssignmentOperator
    right_hand_side: SolcExpressionUnion
    # optional
    argument_types: Optional[List[TypeDescriptionsModel]]


class SolcBinaryOperation(SolidityNode):
    # override alias
    node_type: Literal["BinaryOperation"] = Field(alias="nodeType")
    # required
    is_constant: StrictBool
    is_l_value: StrictBool
    is_pure: StrictBool
    l_value_requested: StrictBool
    type_descriptions: TypeDescriptionsModel
    common_type: TypeDescriptionsModel
    left_expression: SolcExpressionUnion
    operator: BinaryOpOperator
    right_expression: SolcExpressionUnion
    # optional
    argument_types: Optional[List[TypeDescriptionsModel]]
    function: Optional[AstNodeId]  # added in 0.8.19


class SolcConditional(SolidityNode):
    # override alias
    node_type: Literal["Conditional"] = Field(alias="nodeType")
    # required
    is_constant: StrictBool
    is_l_value: StrictBool
    is_pure: StrictBool
    l_value_requested: StrictBool
    type_descriptions: TypeDescriptionsModel
    condition: SolcExpressionUnion
    false_expression: SolcExpressionUnion
    true_expression: SolcExpressionUnion
    # optional
    argument_types: Optional[List[TypeDescriptionsModel]]


class SolcElementaryTypeNameExpression(SolidityNode):
    # override alias
    node_type: Literal["ElementaryTypeNameExpression"] = Field(alias="nodeType")
    # required
    is_constant: StrictBool
    is_l_value: StrictBool
    is_pure: StrictBool
    l_value_requested: StrictBool
    type_descriptions: TypeDescriptionsModel
    type_name: "SolcElementaryTypeName"  # in versions < 0.6.0 this was a string
    # optional
    argument_types: Optional[List[TypeDescriptionsModel]]


class SolcFunctionCall(SolidityNode):
    # override alias
    node_type: Literal["FunctionCall"] = Field(alias="nodeType")
    # required
    is_constant: StrictBool
    is_l_value: StrictBool
    is_pure: StrictBool
    l_value_requested: StrictBool
    type_descriptions: TypeDescriptionsModel
    arguments: List[SolcExpressionUnion]
    expression: SolcExpressionUnion
    kind: FunctionCallKind
    names: List[StrictStr]
    try_call: StrictBool
    # optional
    argument_types: Optional[List[TypeDescriptionsModel]]
    name_locations: Optional[List[Src]]  # added in 0.8.16


class SolcFunctionCallOptions(SolidityNode):
    # override alias
    node_type: Literal["FunctionCallOptions"] = Field(alias="nodeType")
    # required
    is_constant: StrictBool
    is_pure: StrictBool
    l_value_requested: StrictBool
    type_descriptions: TypeDescriptionsModel
    expression: SolcExpressionUnion
    names: List[StrictStr]
    options: List[SolcExpressionUnion]
    # optional
    # TODO:
    is_l_value: Optional[StrictBool]
    argument_types: Optional[List[TypeDescriptionsModel]]


class SolcIdentifier(SolidityNode):
    # override alias
    node_type: Literal["Identifier"] = Field(alias="nodeType")
    # required
    name: StrictStr
    overloaded_declarations: List[AstNodeId]
    type_descriptions: TypeDescriptionsModel
    # optional
    argument_types: Optional[List[TypeDescriptionsModel]]
    referenced_declaration: Optional[AstNodeId]


class SolcIndexAccess(SolidityNode):
    # override alias
    node_type: Literal["IndexAccess"] = Field(alias="nodeType")
    # required
    is_constant: StrictBool
    is_l_value: StrictBool
    is_pure: StrictBool
    l_value_requested: StrictBool
    type_descriptions: TypeDescriptionsModel
    base_expression: SolcExpressionUnion
    # optional
    argument_types: Optional[List[TypeDescriptionsModel]]
    index_expression: OptionalSolcExpressionUnion  # example when this can be None: `abi.decode(params, (address[], uint256))`


class SolcIndexRangeAccess(SolidityNode):
    # override alias
    node_type: Literal["IndexRangeAccess"] = Field(alias="nodeType")
    # required
    is_constant: StrictBool
    is_l_value: StrictBool
    is_pure: StrictBool
    l_value_requested: StrictBool
    type_descriptions: TypeDescriptionsModel
    base_expression: SolcExpressionUnion
    # optional
    argument_types: Optional[List[TypeDescriptionsModel]]
    end_expression: OptionalSolcExpressionUnion
    start_expression: OptionalSolcExpressionUnion


class SolcLiteral(SolidityNode):
    # override alias
    node_type: Literal["Literal"] = Field(alias="nodeType")
    # required
    is_constant: StrictBool
    is_l_value: StrictBool
    is_pure: StrictBool
    l_value_requested: StrictBool
    type_descriptions: TypeDescriptionsModel
    hex_value: StrictStr
    kind: LiteralKind  # hexString new in 0.7.0, prior to 0.7.0 hex strings were marked as strings
    # optional
    argument_types: Optional[List[TypeDescriptionsModel]]
    subdenomination: Optional[StrictStr]  # can be for example "days" or "ether"
    value: Optional[StrictStr]


class SolcMemberAccess(SolidityNode):
    # override alias
    node_type: Literal["MemberAccess"] = Field(alias="nodeType")
    # required
    is_constant: StrictBool
    is_l_value: StrictBool
    is_pure: StrictBool
    l_value_requested: StrictBool
    type_descriptions: TypeDescriptionsModel
    expression: SolcExpressionUnion
    member_name: StrictStr
    # optional
    argument_types: Optional[List[TypeDescriptionsModel]]
    referenced_declaration: Optional[
        AstNodeId
    ]  # because of a bug this is None for enum value access in versions prior to 0.8.2
    member_location: Optional[Src]  # added in 0.8.16


class SolcNewExpression(SolidityNode):
    # override alias
    node_type: Literal["NewExpression"] = Field(alias="nodeType")
    # required
    is_constant: StrictBool
    is_pure: StrictBool
    l_value_requested: StrictBool
    type_descriptions: TypeDescriptionsModel
    type_name: SolcTypeNameUnion
    # optional
    argument_types: Optional[List[TypeDescriptionsModel]]
    is_l_value: Optional[StrictBool]


class SolcTupleExpression(SolidityNode):
    # override alias
    node_type: Literal["TupleExpression"] = Field(alias="nodeType")
    # required
    is_constant: StrictBool
    is_l_value: StrictBool
    is_pure: StrictBool
    l_value_requested: StrictBool
    type_descriptions: TypeDescriptionsModel
    components: List[OptionalSolcExpressionUnion]
    is_inline_array: StrictBool
    # optional
    argument_types: Optional[List[TypeDescriptionsModel]]


class SolcUnaryOperation(SolidityNode):
    # override alias
    node_type: Literal["UnaryOperation"] = Field(alias="nodeType")
    # required
    is_constant: StrictBool
    is_l_value: StrictBool
    is_pure: StrictBool
    l_value_requested: StrictBool
    type_descriptions: TypeDescriptionsModel
    operator: UnaryOpOperator
    prefix: StrictBool
    sub_expression: SolcExpressionUnion
    # optional
    argument_types: Optional[List[TypeDescriptionsModel]]
    function: Optional[AstNodeId]  # added in 0.8.19


class SolcOverrideSpecifier(SolidityNode):
    # override alias
    node_type: Literal["OverrideSpecifier"] = Field(alias="nodeType")
    # required
    overrides: List[Override]
    # optional


# new in 0.8.0 to replace SolcUserDefinedTypeName in many places
class SolcIdentifierPath(SolidityNode):
    node_type: Literal["IdentifierPath"] = Field(alias="nodeType")
    # required
    name: StrictStr
    referenced_declaration: AstNodeId
    # optional
    name_locations: Optional[List[Src]]  # added in 0.8.16


class SolcParameterList(SolidityNode):
    # override alias
    node_type: Literal["ParameterList"] = Field(alias="nodeType")
    # required
    parameters: List["SolcVariableDeclaration"]
    # optional


class SolcTryCatchClause(SolidityNode):
    # override alias
    node_type: Literal["TryCatchClause"] = Field(alias="nodeType")
    # required
    block: "SolcBlock"
    error_name: StrictStr
    # optional
    parameters: Optional["SolcParameterList"]


class SolcStructuredDocumentation(SolidityNode):
    # override alias
    node_type: Literal["StructuredDocumentation"] = Field(alias="nodeType")
    # required
    text: StrictStr
    # optional


class SolcEnumValue(SolidityNode):
    # override alias
    node_type: Literal["EnumValue"] = Field(alias="nodeType")
    # required
    name: StrictStr
    # optional
    name_location: Optional[Src]  # new in 0.8.2


class SolcInheritanceSpecifier(SolidityNode):
    # override alias
    node_type: Literal["InheritanceSpecifier"] = Field(alias="nodeType")
    # required
    base_name: BaseName
    # optional
    arguments: Optional[List[SolcExpressionUnion]]


class SolcModifierInvocation(SolidityNode):
    # override alias
    node_type: Literal["ModifierInvocation"] = Field(alias="nodeType")
    # required
    modifier_name: ModifierName
    # optional
    arguments: Optional[List[SolcExpressionUnion]]
    kind: Optional[
        ModifierInvocationKind
    ]  # new in 0.8.3, fixed in 0.8.4 for base constructor calls


class SolcYulAssignment(YulNode):
    # override alias
    node_type: Literal["YulAssignment"] = Field(alias="nodeType")
    # required
    value: "SolcYulExpressionUnion"
    variable_names: List["SolcYulIdentifier"]
    # optional


class SolcYulBlock(YulNode):
    # override alias
    node_type: Literal["YulBlock"] = Field(alias="nodeType")
    # required
    statements: List["SolcYulStatementUnion"]
    # optional


class SolcYulBreak(YulNode):
    # override alias
    node_type: Literal["YulBreak"] = Field(alias="nodeType")
    # required
    # optional


class SolcYulContinue(YulNode):
    # override alias
    node_type: Literal["YulContinue"] = Field(alias="nodeType")
    # required
    # optional


class SolcYulExpressionStatement(YulNode):
    # override alias
    node_type: Literal["YulExpressionStatement"] = Field(alias="nodeType")
    # required
    expression: "SolcYulExpressionUnion"
    # optional


class SolcYulLeave(YulNode):
    # override alias
    node_type: Literal["YulLeave"] = Field(alias="nodeType")
    # required
    # optional


class SolcYulForLoop(YulNode):
    # override alias
    node_type: Literal["YulForLoop"] = Field(alias="nodeType")
    # required
    body: "SolcYulBlock"
    condition: "SolcYulExpressionUnion"
    post: "SolcYulBlock"
    pre: "SolcYulBlock"
    # optional


class SolcYulFunctionDefinition(YulNode):
    # override alias
    node_type: Literal["YulFunctionDefinition"] = Field(alias="nodeType")
    # required
    body: "SolcYulBlock"
    name: StrictStr
    # optional
    parameters: Optional[List["SolcYulTypedName"]]
    return_variables: Optional[List["SolcYulTypedName"]]


class SolcYulIf(YulNode):
    # override alias
    node_type: Literal["YulIf"] = Field(alias="nodeType")
    # required
    body: "SolcYulBlock"
    condition: "SolcYulExpressionUnion"
    # optional


class SolcYulSwitch(YulNode):
    # override alias
    node_type: Literal["YulSwitch"] = Field(alias="nodeType")
    # required
    cases: List["SolcYulCase"]
    expression: "SolcYulExpressionUnion"
    # optional


class SolcYulVariableDeclaration(YulNode):
    # override alias
    node_type: Literal["YulVariableDeclaration"] = Field(alias="nodeType")
    # required
    variables: List["SolcYulTypedName"]
    # optional
    value: OptionalSolcYulExpressionUnion


class SolcYulFunctionCall(YulNode):
    # override alias
    node_type: Literal["YulFunctionCall"] = Field(alias="nodeType")
    # required
    arguments: List[SolcYulExpressionUnion]
    function_name: "SolcYulIdentifier"
    # optional


class SolcYulIdentifier(YulNode):
    # override alias
    node_type: Literal["YulIdentifier"] = Field(alias="nodeType")
    # required
    name: StrictStr
    # optional


class SolcYulLiteral(YulNode):
    # override alias
    node_type: Literal["YulLiteral"] = Field(alias="nodeType")
    # required
    kind: YulLiteralValueKind
    type: StrictStr
    # at least one of these should be set
    value: Optional[StrictStr]
    hex_value: Optional[StrictStr]  # sice 0.8.5

    @root_validator
    def value_or_hex_value_set(cls, values):
        value, hex_value = values.get("value"), values.get("hex_value")
        assert (
            value is not None or hex_value is not None
        ), "YulLiteral: either 'value' or 'hex_value' must be set"
        return values


class SolcYulTypedName(YulNode):
    # override alias
    node_type: Literal["YulTypedName"] = Field(alias="nodeType")
    # required
    name: StrictStr
    type: StrictStr
    # optional


class SolcYulCase(YulNode):
    # override alias
    node_type: Literal["YulCase"] = Field(alias="nodeType")
    # required
    body: "SolcYulBlock"
    value: Union[Literal["default"], SolcYulLiteral]
    # optional


# region update_forward_refs
SolcSourceUnit.update_forward_refs()
SolcPragmaDirective.update_forward_refs()
SolcImportDirective.update_forward_refs()
SolcVariableDeclaration.update_forward_refs()
SolcEnumDefinition.update_forward_refs()
SolcFunctionDefinition.update_forward_refs()
SolcStructDefinition.update_forward_refs()
SolcErrorDefinition.update_forward_refs()
SolcUserDefinedValueTypeDefinition.update_forward_refs()
SolcContractDefinition.update_forward_refs()
SolcEventDefinition.update_forward_refs()
SolcModifierDefinition.update_forward_refs()
UsingForDirectiveFunction.update_forward_refs()
SolcUsingForDirective.update_forward_refs()
SolcArrayTypeName.update_forward_refs()
SolcElementaryTypeName.update_forward_refs()
SolcFunctionTypeName.update_forward_refs()
SolcMapping.update_forward_refs()
SolcUserDefinedTypeName.update_forward_refs()
SolcBlock.update_forward_refs()
SolcBreak.update_forward_refs()
SolcContinue.update_forward_refs()
SolcDoWhileStatement.update_forward_refs()
SolcEmitStatement.update_forward_refs()
SolcExpressionStatement.update_forward_refs()
SolcForStatement.update_forward_refs()
SolcIfStatement.update_forward_refs()
SolcInlineAssembly.update_forward_refs()
SolcPlaceholderStatement.update_forward_refs()
SolcReturn.update_forward_refs()
SolcRevertStatement.update_forward_refs()
SolcTryStatement.update_forward_refs()
SolcUncheckedBlock.update_forward_refs()
SolcVariableDeclarationStatement.update_forward_refs()
SolcWhileStatement.update_forward_refs()
SolcAssignment.update_forward_refs()
SolcBinaryOperation.update_forward_refs()
SolcConditional.update_forward_refs()
SolcElementaryTypeNameExpression.update_forward_refs()
SolcFunctionCall.update_forward_refs()
SolcFunctionCallOptions.update_forward_refs()
SolcIdentifier.update_forward_refs()
SolcIndexAccess.update_forward_refs()
SolcIndexRangeAccess.update_forward_refs()
SolcLiteral.update_forward_refs()
SolcMemberAccess.update_forward_refs()
SolcNewExpression.update_forward_refs()
SolcTupleExpression.update_forward_refs()
SolcUnaryOperation.update_forward_refs()
SolcOverrideSpecifier.update_forward_refs()
SolcIdentifierPath.update_forward_refs()
SolcParameterList.update_forward_refs()
SolcTryCatchClause.update_forward_refs()
SolcStructuredDocumentation.update_forward_refs()
SolcEnumValue.update_forward_refs()
SymbolAliasModel.update_forward_refs()
SolcYulAssignment.update_forward_refs()
SolcYulBlock.update_forward_refs()
SolcYulBreak.update_forward_refs()
SolcYulContinue.update_forward_refs()
SolcYulExpressionStatement.update_forward_refs()
SolcYulLeave.update_forward_refs()
SolcYulForLoop.update_forward_refs()
SolcYulFunctionDefinition.update_forward_refs()
SolcYulIf.update_forward_refs()
SolcYulSwitch.update_forward_refs()
SolcYulVariableDeclaration.update_forward_refs()
SolcYulFunctionCall.update_forward_refs()
SolcYulIdentifier.update_forward_refs()
SolcYulLiteral.update_forward_refs()
SolcYulTypedName.update_forward_refs()
SolcYulCase.update_forward_refs()
# endregion update_forward_refs



from __future__ import annotations

from typing import List, Set

import woke.ir as ir
from woke.detectors import (
    Detection,
    DetectionConfidence,
    DetectionImpact,
    Detector,
    DetectorResult,
    detector,
)


class FunctionCallOptionsNotCalledDetector(Detector):
    _detections: Set[DetectorResult]

    def __init__(self):
        self._detections = set()

    def detect(self) -> List[DetectorResult]:
        return list(self._detections)

    def visit_function_call_options(self, node: ir.FunctionCallOptions):
        if not isinstance(node.parent, ir.FunctionCall):
            self._detections.add(
                DetectorResult(
                    Detection(node, "Function call options not called"),
                    DetectionImpact.HIGH,
                    DetectionConfidence.HIGH,
                )
            )

    def visit_member_access(self, node: ir.MemberAccess):
        if node.referenced_declaration in {
            ir.enums.GlobalSymbolsEnum.FUNCTION_GAS,
            ir.enums.GlobalSymbolsEnum.FUNCTION_VALUE,
        }:
            parent = node.parent
            assert isinstance(parent, ir.FunctionCall)

            if not isinstance(parent.parent, ir.FunctionCall):
                if (
                    node.referenced_declaration
                    == ir.enums.GlobalSymbolsEnum.FUNCTION_GAS
                ):
                    self._detections.add(
                        DetectorResult(
                            Detection(node, "Function with gas not called"),
                            DetectionImpact.HIGH,
                            DetectionConfidence.HIGH,
                        )
                    )
                else:
                    self._detections.add(
                        DetectorResult(
                            Detection(node, "Function with value not called"),
                            DetectionImpact.HIGH,
                            DetectionConfidence.HIGH,
                        )
                    )

    @detector.command("function-call-options-not-called")
    def cli(self):
        """
        Function with gas or value set actually is not called, e.g. `this.externalFunction.value(targetValue)` or `this.externalFunction{value: targetValue}`.
        """
        pass



from __future__ import annotations

from typing import List, Set

import woke.ir as ir
import woke.ir.types as types
from woke.detectors import (
    Detection,
    DetectionConfidence,
    DetectionImpact,
    Detector,
    DetectorResult,
    detector,
)


class UncheckedReturnValueDetector(Detector):
    _detections: Set[DetectorResult]

    def __init__(self):
        self._detections = set()

    def detect(self) -> List[DetectorResult]:
        return list(self._detections)

    def visit_function_call(self, node: ir.FunctionCall):
        t = node.type
        has_return_value = not (isinstance(t, types.Tuple) and len(t.components) == 0)
        if not has_return_value:
            return

        # TODO external call in try statement

        current_node = node
        nodes = []
        is_expression_statement = False
        while current_node is not None:
            if isinstance(current_node, ir.ExpressionStatement):
                is_expression_statement = True
                break
            nodes.append(current_node)
            current_node = current_node.parent

            if isinstance(
                current_node, (ir.Assignment, ir.FunctionCall, ir.FunctionCallOptions)
            ):
                return
            elif isinstance(current_node, ir.Conditional):
                if current_node.condition == nodes[-1]:
                    return
            elif isinstance(current_node, ir.IndexAccess):
                if current_node.index_expression == nodes[-1]:
                    return
            elif isinstance(current_node, ir.IndexRangeAccess):
                if (
                    current_node.start_expression == nodes[-1]
                    or current_node.end_expression == nodes[-1]
                ):
                    return
            elif isinstance(current_node, ir.UnaryOperation):
                if current_node.operator in {
                    ir.enums.UnaryOpOperator.PLUS_PLUS,
                    ir.enums.UnaryOpOperator.MINUS_MINUS,
                    ir.enums.UnaryOpOperator.DELETE,
                }:
                    return

        if not is_expression_statement:
            return

        self._detections.add(
            DetectorResult(
                Detection(node, "Unchecked return value"),
                impact=DetectionImpact.MEDIUM,
                confidence=DetectionConfidence.HIGH,
            )
        )

    @detector.command("unchecked-return-value")
    def cli(self):
        """
        Return value of a function call is ignored.
        """
        pass



from __future__ import annotations

import logging
from typing import List, Sequence, Set

import woke.ir as ir
import woke.ir.types as types
from woke.core.solidity_version import SolidityVersionRange, SolidityVersionRanges
from woke.detectors import (
    Detection,
    DetectionConfidence,
    DetectionImpact,
    Detector,
    DetectorResult,
    detector,
)

logger = logging.getLogger(__name__)


def type_contains_dynamic_component(t: types.TypeAbc) -> bool:

    if isinstance(t, types.Array):
        if t.length is None:
            return True
        return type_contains_dynamic_component(t.base_type)
    elif isinstance(t, (types.Bytes, types.String)):
        return True
    elif isinstance(t, types.Tuple):
        return any(
            type_contains_dynamic_component(component)
            for component in t.components
            if component is not None
        )
    elif isinstance(t, types.Struct):
        return any(type_contains_dynamic_component(m.type) for m in t.ir_node.members)
    else:
        return False


def abi_encoder_v2_enabled(node: ir.IrAbc) -> bool:

    source_unit = node
    while source_unit is not None:
        if isinstance(source_unit, ir.SourceUnit):
            break
        source_unit = source_unit.parent
    assert isinstance(source_unit, ir.SourceUnit)

    if any(
        directive.literals == ("abicoder", "v1") for directive in source_unit.pragmas
    ):
        return False
    elif any(
        directive.literals == ("abicoder", "v2") for directive in source_unit.pragmas
    ):
        return True
    elif any(
        directive.literals == ("experimental", "ABIEncoderV2")
        for directive in source_unit.pragmas
    ):
        return True
    else:
        default_v2_versions = SolidityVersionRanges(
            [SolidityVersionRange("0.8.0", True, None, None)]
        )
        if len(default_v2_versions & node.version_ranges) == 0:
            return False
        return True


def types_meet_requirements(
    t: Sequence[types.TypeAbc], data_location_is_calldata: bool = False
) -> bool:
    for component in t:
        if isinstance(component, types.Struct):
            bug = types_meet_requirements(
                list(m.type for m in component.ir_node.members),
                data_location_is_calldata
                or component.data_location == ir.enums.DataLocation.CALLDATA,
            )
            if bug:
                return True

    if len(t) < 2:
        return False
    last = t[-1]
    if not (
        isinstance(last, types.Array)
        and last.length is not None
        and (
            last.data_location == ir.enums.DataLocation.CALLDATA
            or data_location_is_calldata
        )
    ):
        return False

    base_type = last.base_type
    while isinstance(base_type, types.Array):
        if base_type.length is None:
            return False
        base_type = base_type.base_type
    if isinstance(base_type, types.UInt) and base_type.bits_count == 256:
        pass
    elif isinstance(base_type, types.FixedBytes) and base_type.bytes_count == 32:
        pass
    else:
        return False

    if not any(type_contains_dynamic_component(x) for x in t[:-1]):
        return False

    return True


class OverflowCalldataTupleReencodingBugDetector(Detector):
    _detections: Set[DetectorResult]

    def __init__(self):
        self._detections = set()

    def detect(self) -> List[DetectorResult]:
        return list(self._detections)

    def visit_function_definition(self, node: ir.FunctionDefinition):
        if node.visibility not in {
            ir.enums.Visibility.PUBLIC,
            ir.enums.Visibility.EXTERNAL,
        }:
            return None

        versions = node.version_ranges
        affected_versions = SolidityVersionRanges(
            [SolidityVersionRange("0.5.8", True, "0.8.16", False)]
        )
        if len(versions & affected_versions) == 0:
            return

        # ABI encoder v2 must be enabled
        if not abi_encoder_v2_enabled(node):
            return

        encoded_types = [param.type for param in node.return_parameters.parameters]
        if types_meet_requirements(encoded_types):
            self._detections.add(
                DetectorResult(
                    Detection(
                        node.return_parameters,
                        "Found head overflow calldata tuple reencoding compiler bug",
                    ),
                    impact=DetectionImpact.HIGH,
                    confidence=DetectionConfidence.HIGH,
                )
            )

    def visit_function_call(self, node: ir.FunctionCall):
        if node.kind == ir.enums.FunctionCallKind.TYPE_CONVERSION:
            return

        versions = node.version_ranges
        affected_versions = SolidityVersionRanges(
            [SolidityVersionRange("0.5.8", True, "0.8.16", False)]
        )
        if len(versions & affected_versions) == 0:
            return

        func_identifier = node.expression
        while True:
            if isinstance(func_identifier, (ir.Identifier, ir.MemberAccess)):
                break
            elif isinstance(func_identifier, ir.FunctionCallOptions):
                func_identifier = func_identifier.expression
            elif isinstance(func_identifier, ir.NewExpression):
                return
            elif (
                isinstance(func_identifier, ir.TupleExpression)
                and len(func_identifier.components) == 1
            ):
                func_identifier = func_identifier.components[0]
            elif isinstance(func_identifier, ir.FunctionCall):
                t = func_identifier.type
                if isinstance(t, types.Function) and (t.value_set or t.gas_set):
                    func_identifier = func_identifier.expression
                else:
                    # function returning a function
                    break
            else:
                logger.warning(
                    f"Unexpected function call child node: {func_identifier} {func_identifier.source if func_identifier else ''}"
                )
                return

        t = func_identifier.type

        if not isinstance(t, types.Function):
            return

        if t.kind == ir.enums.FunctionTypeKind.EXTERNAL:
            assert t.attached_to is None
            encoded_types = [arg.type for arg in node.arguments]
        elif t.kind in {
            ir.enums.FunctionTypeKind.ERROR,
            ir.enums.FunctionTypeKind.EVENT,
        }:
            encoded_types = [arg.type for arg in node.arguments]
        elif t.kind in {
            ir.enums.FunctionTypeKind.ABI_ENCODE,
            ir.enums.FunctionTypeKind.ABI_ENCODE_PACKED,
        }:
            encoded_types = [arg.type for arg in node.arguments]
        elif t.kind in {
            ir.enums.FunctionTypeKind.ABI_ENCODE_WITH_SELECTOR,
            ir.enums.FunctionTypeKind.ABI_ENCODE_WITH_SIGNATURE,
        }:
            assert len(node.arguments) >= 1
            encoded_types = [arg.type for arg in node.arguments[1:]]
        elif t.kind == ir.enums.FunctionTypeKind.ABI_ENCODE_CALL:
            assert len(node.arguments) == 2
            if isinstance(node.arguments[1], ir.TupleExpression):
                encoded_types = [
                    arg.type if arg is not None else None
                    for arg in node.arguments[
                        1
                    ].components  # pyright: ignore reportGeneralTypeIssues
                ]
            elif (
                isinstance(node.arguments[1], ir.FunctionCall)
                and node.arguments[1].kind
                == ir.enums.FunctionCallKind.STRUCT_CONSTRUCTOR_CALL
            ):
                # probably always created in memory, not calldata
                return
            else:
                logger.warning(
                    f"Unexpected abi.encodeCall second argument: {node.arguments[1]}"
                )
                return
        else:
            return

        # ABI encoder v2 must be enabled
        if not abi_encoder_v2_enabled(node):
            return

        if types_meet_requirements(
            encoded_types  # pyright: ignore reportGeneralTypeIssues
        ):
            self._detections.add(
                DetectorResult(
                    Detection(
                        node,
                        "Found head overflow calldata tuple reencoding compiler bug",
                    ),
                    impact=DetectionImpact.HIGH,
                    confidence=DetectionConfidence.HIGH,
                )
            )

    @detector.command("overflow-calldata-tuple-reencoding-bug")
    def cli(self):
        """
        Detects Head Overflow Calldata Tuple Reencoding compiler bug
        """
        pass



from typing import Dict, List, Optional, Tuple

from pydantic import BaseModel, Extra

from woke.utils import StrEnum

__doc__ = """Solc standard JSON output data model as described by https://docs.soliditylang.org/en/v0.8.12/using-the-compiler.html#output-description"""


def _to_camel(s: str) -> str:
    split = s.split("_")
    return split[0].lower() + "".join([w.capitalize() for w in split[1:]])


class SolcOutputModel(BaseModel):
    class Config:
        alias_generator = _to_camel
        extra = Extra.forbid


class SolcOutputErrorSourceLocation(SolcOutputModel):
    file: str
    start: int
    end: int

    def __members(self) -> Tuple:
        return self.file, self.start, self.end

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__members() == other.__members()
        return NotImplemented

    def __hash__(self):
        return hash(self.__members())


class SolcOutputErrorSecondarySourceLocation(SolcOutputModel):
    file: Optional[str]
    start: Optional[int]
    end: Optional[int]
    message: str

    def __members(self) -> Tuple:
        return self.file, self.start, self.end, self.message

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__members() == other.__members()
        return NotImplemented

    def __hash__(self):
        return hash(self.__members())


class SolcOutputErrorTypeEnum(StrEnum):
    """Solc output error types as described by https://docs.soliditylang.org/en/v0.8.11/using-the-compiler.html#error-types"""

    JSON_ERROR = "JSONError"
    """JSON input doesn’t conform to the required format, e.g. input is not a JSON object, the language is not supported, etc."""
    IO_ERROR = "IOError"
    """IO and import processing errors, such as unresolvable URL or hash mismatch in supplied sources."""
    PARSER_ERROR = "ParserError"
    """Source code doesn’t conform to the language rules."""
    DOCSTRING_PARSING_ERROR = "DocstringParsingError"
    """The NatSpec tags in the comment block cannot be parsed."""
    SYNTAX_ERROR = "SyntaxError"
    """Syntactical error, such as continue is used outside of a for loop."""
    DECLARATION_ERROR = "DeclarationError"
    """Invalid, unresolvable or clashing identifier names. e.g. `Identifier not found`"""
    TYPE_ERROR = "TypeError"
    """Error within the type system, such as invalid type conversions, invalid assignments, etc."""
    UNIMPLEMENTED_FEATURE_ERROR = "UnimplementedFeatureError"
    """Feature is not supported by the compiler, but is expected to be supported in future versions."""
    INTERNAL_COMPILER_ERROR = "InternalCompilerError"
    """Internal bug triggered in the compiler - this should be reported as an issue."""
    EXCEPTION = "Exception"
    """Unknown failure during compilation - this should be reported as an issue."""
    COMPILER_ERROR = "CompilerError"
    """Invalid use of the compiler stack - this should be reported as an issue."""
    FATAL_ERROR = "FatalError"
    """Fatal error not processed correctly - this should be reported as an issue."""
    WARNING = "Warning"
    """A warning, which didn’t stop the compilation, but should be addressed if possible."""
    INFO = "Info"
    """Information that the compiler thinks the user might find useful, but is not dangerous and does not necessarily need to be addressed."""
    YUL_EXCEPTION = "YulException"
    CODE_GENERATION_ERROR = "CodeGenerationError"


class SolcOutputErrorSeverityEnum(StrEnum):
    ERROR = "error"
    WARNING = "warning"
    INFO = "info"


class SolcOutputStorageLayoutTypeEncodingEnum(StrEnum):
    INPLACE = "inplace"
    """Data is laid out contiguously in storage"""
    MAPPING = "mapping"
    """Keccak-256 hash-based method (see https://docs.soliditylang.org/en/latest/internals/layout_in_storage.html#mappings-and-dynamic-arrays)"""
    DYNAMIC_ARRAY = "dynamic_array"
    """Keccak-256 hash-based method (see https://docs.soliditylang.org/en/latest/internals/layout_in_storage.html#mappings-and-dynamic-arrays)"""
    BYTES = "bytes"
    """Single slot or Keccak-256 hash-based depending on the data size (see https://docs.soliditylang.org/en/latest/internals/layout_in_storage.html#bytes-and-string)"""


class SolcOutputError(SolcOutputModel):
    source_location: Optional[SolcOutputErrorSourceLocation]
    secondary_source_locations: Optional[List[SolcOutputErrorSecondarySourceLocation]]
    type: SolcOutputErrorTypeEnum
    component: str
    severity: SolcOutputErrorSeverityEnum
    error_code: Optional[str]
    message: str
    formatted_message: Optional[str]

    def __members(self) -> Tuple:
        return (
            self.source_location,
            tuple(self.secondary_source_locations)
            if self.secondary_source_locations
            else None,
            self.type,
            self.component,
            self.severity,
            self.error_code,
            self.message,
            self.formatted_message,
        )

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__members() == other.__members()
        return NotImplemented

    def __hash__(self):
        return hash(self.__members())


class SolcOutputSourceInfo(SolcOutputModel):
    id: int
    ast: Optional[Dict]


class SolcOutputStorageLayoutStorage(SolcOutputModel):
    ast_id: int
    contract: str
    label: str
    offset: int
    slot: int
    type: str


class SolcOutputStorageLayoutType(SolcOutputModel):
    encoding: SolcOutputStorageLayoutTypeEncodingEnum
    label: str
    number_of_bytes: int
    base: Optional[str] = None
    key: Optional[str] = None
    value: Optional[str] = None
    members: Optional[List[SolcOutputStorageLayoutStorage]] = None


class SolcOutputStorageLayout(SolcOutputModel):
    storage: List[SolcOutputStorageLayoutStorage]
    types: Optional[Dict[str, SolcOutputStorageLayoutType]] = None


class SolcOutputEvmBytecodeFunctionDebugData(SolcOutputModel):
    entry_point: Optional[int]
    id: Optional[int]
    parameter_slots: int
    return_slots: int


class SolcOutputEvmBytecodeGeneratedSources(SolcOutputModel):
    ast: Dict
    contents: str
    id: int
    language: str
    name: str


class SolcOutputEvmBytecodeLinkReferencesInfo(SolcOutputModel):
    start: int
    length: int


class SolcOutputEvmBytecodeData(SolcOutputModel):
    function_debug_data: Optional[
        Dict[str, SolcOutputEvmBytecodeFunctionDebugData]
    ] = None  # internal name of the function -> debug data
    object: Optional[str] = None
    opcodes: Optional[str] = None
    source_map: Optional[str] = None
    generated_sources: Optional[List[SolcOutputEvmBytecodeGeneratedSources]] = None
    link_references: Optional[
        Dict[str, Dict[str, List[SolcOutputEvmBytecodeLinkReferencesInfo]]]
    ] = None


class SolcOutputEvmDeployedBytecodeData(SolcOutputEvmBytecodeData):
    immutable_references: Optional[
        Dict[str, List[SolcOutputEvmBytecodeLinkReferencesInfo]]
    ] = None


class SolcOutputEvmData(SolcOutputModel):
    assembly: Optional[str] = None
    legacy_assembly: Optional[Dict] = None  # TODO
    bytecode: Optional[SolcOutputEvmBytecodeData] = None
    deployed_bytecode: Optional[SolcOutputEvmDeployedBytecodeData] = None
    method_identifiers: Optional[Dict[str, str]] = None
    gas_estimates: Optional[Dict] = None  # TODO


class SolcOutputEwasmData(SolcOutputModel):
    wast: Optional[str]
    """S-expressions format"""
    wasm: Optional[str]
    """Binary format (hex string)"""


# TODO provide better data model for solc output
class SolcOutputContractInfo(SolcOutputModel):
    abi: Optional[List] = None
    """The Ethereum Contract ABI. If empty, it is represented as an empty array. See https://docs.soliditylang.org/en/latest/abi-spec.html."""
    metadata: Optional[str] = None
    """Serialized JSON metadata string. See https://docs.soliditylang.org/en/latest/metadata.html."""
    userdoc: Optional[Dict] = None
    """User documentation (natspec)"""
    devdoc: Optional[Dict] = None
    """Developer documentation (natspec)"""
    ir: Optional[str] = None
    """Intermediate representation (string)"""
    storage_layout: Optional[SolcOutputStorageLayout] = None
    """See https://docs.soliditylang.org/en/latest/internals/layout_in_storage.html#json-output"""
    evm: Optional[SolcOutputEvmData] = None
    """EVM-related outputs"""
    ewasm: Optional[SolcOutputEwasmData] = None
    """Ewasm related outputs"""


class SolcOutput(SolcOutputModel):
    errors: List[SolcOutputError] = []
    sources: Dict[str, SolcOutputSourceInfo] = {}
    contracts: Dict[
        str, Dict[str, SolcOutputContractInfo]
    ] = {}  # source_unit_name -> (contract_name -> info)



import logging
from pathlib import Path
from typing import List, Optional, Tuple, Union

from woke.ir import (
    BinaryOperation,
    DeclarationAbc,
    FunctionDefinition,
    Identifier,
    IdentifierPath,
    IrAbc,
    MemberAccess,
    ModifierDefinition,
    SourceUnit,
    UnaryOperation,
    UserDefinedTypeName,
    VariableDeclaration,
    YulIdentifier,
)
from woke.ir.enums import GlobalSymbolsEnum
from woke.lsp.common_structures import (
    DocumentUri,
    Location,
    LocationLink,
    PartialResultParams,
    Position,
    TextDocumentPositionParams,
    TextDocumentRegistrationOptions,
    WorkDoneProgressOptions,
    WorkDoneProgressParams,
)
from woke.lsp.context import LspContext
from woke.lsp.lsp_compiler import LspCompiler
from woke.lsp.utils import path_to_uri, position_within_range, uri_to_path
from woke.lsp.utils.position import changes_to_byte_offset

logger = logging.getLogger(__name__)


class DefinitionOptions(WorkDoneProgressOptions):
    pass


class DefinitionRegistrationOptions(TextDocumentRegistrationOptions, DefinitionOptions):
    pass


class DefinitionParams(
    TextDocumentPositionParams, WorkDoneProgressParams, PartialResultParams
):
    pass


def _create_location(
    path: Path, location: Tuple[int, int], compiler: LspCompiler
) -> Location:
    return Location(
        uri=DocumentUri(path_to_uri(path)),
        range=compiler.get_range_from_byte_offsets(path, location),
    )


def _get_results_from_node(
    original_node: Union[IrAbc, GlobalSymbolsEnum],
    position: Position,
    context: LspContext,
    byte_offset: int,
    node_name_location: Optional[Tuple[int, int]],
) -> Optional[List[Tuple[Path, Tuple[int, int]]]]:
    if isinstance(original_node, DeclarationAbc):
        assert node_name_location is not None
        name_location_range = context.compiler.get_range_from_byte_offsets(
            original_node.file, node_name_location
        )
        if position_within_range(position, name_location_range):
            return [(original_node.file, original_node.name_location)]
        return None

    if isinstance(original_node, (Identifier, MemberAccess)):
        node = original_node.referenced_declaration
        if node is None:
            return None
    elif isinstance(original_node, (IdentifierPath, UserDefinedTypeName)):
        part = original_node.identifier_path_part_at(byte_offset)
        if part is None:
            return None
        node = part.referenced_declaration
    elif isinstance(original_node, YulIdentifier):
        external_reference = original_node.external_reference
        if external_reference is None:
            return None
        node = external_reference.referenced_declaration
    elif (
        isinstance(original_node, (UnaryOperation, BinaryOperation))
        and original_node.function is not None
    ):
        node = original_node.function
    else:
        node = original_node

    if not isinstance(node, (DeclarationAbc, SourceUnit)):
        return None

    definitions = []

    if isinstance(node, (FunctionDefinition, VariableDeclaration)):
        if isinstance(node, VariableDeclaration) or node.implemented:
            definitions.append((node.file, node.name_location))

        for base_function in node.base_functions:
            if base_function.implemented:
                definitions.append((base_function.file, base_function.name_location))
        if isinstance(node, FunctionDefinition):
            for child_function in node.child_functions:
                if isinstance(child_function, VariableDeclaration):
                    definitions.append(
                        (child_function.file, child_function.name_location)
                    )
                elif (
                    isinstance(child_function, FunctionDefinition)
                    and child_function.implemented
                ):
                    definitions.append(
                        (child_function.file, child_function.name_location)
                    )
    elif isinstance(node, ModifierDefinition):
        if node.implemented:
            definitions.append((node.file, node.name_location))
        for base_modifier in node.base_modifiers:
            if base_modifier.implemented:
                definitions.append((base_modifier.file, base_modifier.name_location))
        for child_modifier in node.child_modifiers:
            if child_modifier.implemented:
                definitions.append((child_modifier.file, child_modifier.name_location))
    elif isinstance(node, SourceUnit):
        definitions.append((node.file, node.byte_location))
    else:
        definitions.append((node.file, node.name_location))

    return definitions


async def _get_definition_from_cache(
    path: Path,
    position: Position,
    context: LspContext,
):
    new_byte_offset = context.compiler.get_byte_offset_from_line_pos(
        path, position.line, position.character
    )
    backward_changes = context.compiler.get_last_compilation_backward_changes(path)
    if backward_changes is None:
        raise Exception("No backward changes found")
    changes_before = backward_changes[0:new_byte_offset]
    old_byte_offset = changes_to_byte_offset(changes_before) + new_byte_offset

    tree = context.compiler.last_compilation_interval_trees[path]

    intervals = tree.at(old_byte_offset)
    nodes: List[IrAbc] = [interval.data for interval in intervals]
    if len(nodes) == 0:
        raise ValueError(f"Could not find node at {old_byte_offset}")

    node = max(nodes, key=lambda n: n.ast_tree_depth)

    if isinstance(node, (DeclarationAbc, SourceUnit)):
        if isinstance(node, DeclarationAbc):
            location = node.name_location
        else:
            location = node.byte_location
        forward_changes = context.compiler.get_last_compilation_forward_changes(path)
        if forward_changes is None:
            raise Exception("No forward changes found")
        new_start = (
            changes_to_byte_offset(forward_changes[0 : location[0]]) + location[0]
        )
        new_end = changes_to_byte_offset(forward_changes[0 : location[1]]) + location[1]
        node_name_location = (new_start, new_end)
    else:
        node_name_location = None

    result = _get_results_from_node(
        node, position, context, old_byte_offset, node_name_location
    )

    if result is None:
        return None

    if len(result) == 1:
        path, location = result[0]
        forward_changes = context.compiler.get_last_compilation_forward_changes(path)
        if forward_changes is None:
            raise Exception("No forward changes found")
        new_start = (
            changes_to_byte_offset(forward_changes[0 : location[0]]) + location[0]
        )
        new_end = changes_to_byte_offset(forward_changes[0 : location[1]]) + location[1]
        return _create_location(path, (new_start, new_end), context.compiler)
    else:
        ret = []
        for path, location in result:
            forward_changes = context.compiler.get_last_compilation_forward_changes(
                path
            )
            if forward_changes is None:
                raise Exception("No forward changes found")
            new_start = (
                changes_to_byte_offset(forward_changes[0 : location[0]]) + location[0]
            )
            new_end = (
                changes_to_byte_offset(forward_changes[0 : location[1]]) + location[1]
            )
            ret.append(_create_location(path, (new_start, new_end), context.compiler))
        return ret


async def definition(
    context: LspContext, params: DefinitionParams
) -> Union[Location, List[Location], List[LocationLink], None]:
    logger.debug(
        f"Requested definition for file {params.text_document.uri} at {params.position}"
    )

    path = uri_to_path(params.text_document.uri).resolve()
    if (
        path not in context.compiler.interval_trees
        or not context.compiler.output_ready.is_set()
    ):
        # try to use old build artifacts
        try:
            return await _get_definition_from_cache(path, params.position, context)
        except Exception:
            pass

    await context.compiler.output_ready.wait()
    if path not in context.compiler.interval_trees:
        return None

    tree = context.compiler.interval_trees[path]

    byte_offset = context.compiler.get_byte_offset_from_line_pos(
        path, params.position.line, params.position.character
    )
    intervals = tree.at(byte_offset)
    nodes: List[IrAbc] = [interval.data for interval in intervals]
    logger.debug(
        f"Found {len(nodes)} nodes at byte offset {byte_offset}:\n{sorted(nodes, key=lambda x: x.ast_tree_depth)}"
    )
    if len(nodes) == 0:
        return None

    node = max(nodes, key=lambda n: n.ast_tree_depth)
    logger.debug(f"Found node {node}")

    if isinstance(node, DeclarationAbc):
        node_name_location = node.name_location
    else:
        node_name_location = None

    result = _get_results_from_node(
        node, params.position, context, byte_offset, node_name_location
    )
    if result is None:
        return None

    if len(result) == 1:
        path, location = result[0]
        return _create_location(path, location, context.compiler)
    else:
        return [
            _create_location(path, location, context.compiler)
            for path, location in result
        ]



from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING

import rich_click as click

if TYPE_CHECKING:
    from woke.config import WokeConfig


logger = logging.getLogger(__name__)


async def run_server(config: WokeConfig, port: int) -> None:
    from woke.lsp.server import LspServer

    async def client_callback(
        reader: asyncio.StreamReader, writer: asyncio.StreamWriter
    ) -> None:
        lsp_server = LspServer(config, reader, writer)
        logger.info("Client connected")
        await lsp_server.run()
        writer.close()
        logger.info("Client disconnected")

    server = await asyncio.start_server(client_callback, port=port)
    logger.info(f"Started LSP server on port {port}")

    async with server:
        await server.serve_forever()


@click.command(name="lsp")
@click.option(
    "--port",
    default=65432,
    type=int,
    help="Port to listen on.",
    show_default=True,
)
@click.pass_context
def run_lsp(context: click.Context, port: int):
    """
    Start the LSP server.
    """
    from woke.config import WokeConfig

    config = WokeConfig()
    config.load_configs()  # load ~/.woke/config.toml and ./woke.toml

    asyncio.run(run_server(config, port))
